import math
import random
from typing import Union

import numpy as np

from ..sweep import Sweep
from ..uncalculated_sweep import UncalculatedSweep
from ..utility_functions.a_t import tukey_a_t
from ..axis import ArrayAxis, get_array_axis_from_array


class SegmentSmallerThenDtError(Exception):
    pass


def get_shuffle(
    time: Union[ArrayAxis, np.ndarray],
    f_start=1.0,
    f_end=101.0,
    length_time_segments=0.5,
    round_number_frequency: int = None,
    time_tapper=1.0,
) -> Sweep:
    """Create shuffle sweep signal.

    t_tapper in seconds is used to apply tukey function at the end of dwell sweep signal.
    """
    if not isinstance(time, ArrayAxis):
        time = get_array_axis_from_array(time)

    if time.sample >= length_time_segments:
        raise SegmentSmallerThenDtError(
            f"Length of shuffle time segment \
            (length_time_segments - {length_time_segments}) should be grate \
             then sample rate (dt - {time.sample})"
        )

    n_segments = math.ceil(time.end / length_time_segments)

    f_segment = np.linspace(f_start, f_end, n_segments + 1)

    if round_number_frequency:
        f_segment = np.round(
            f_segment, round_number_frequency
        )

    random.shuffle(f_segment)

    x = np.linspace(
        0.0,
        length_time_segments,
        math.ceil(length_time_segments / time.sample) + 1
    )
    my_cos = np.cos(x * np.pi / length_time_segments - np.pi) / 2 + 1 / 2
    f_t = np.array([], dtype='float')
    for f1, f2 in zip(f_segment[:-1], f_segment[1:]):
        f_t = np.append(f_t, (my_cos * (f2 - f1) + f1)[1:])

    a_t = tukey_a_t(time.array, time_tapper)
    uncalculated_sweep = UncalculatedSweep(time, f_t[:-2], a_t)

    return uncalculated_sweep()



from copy import copy
from typing import Optional

import numpy as np

from .help_types import ArrayLike, RealNumber


class ArrayAxis:

    def __init__(self, start: RealNumber, end: RealNumber,
                 sample: RealNumber) -> None:
        '''The representation of some array axis.

        This is description of sequence. The start sequence. The end sequence.
        And the sample spacing between elements.

        Args:
            start (RealNumber): is start of array.
            end (RealNumber): is end of array.
            sample (RealNumber): is sample of array.
        '''
        self._start = start
        self._end = end
        self._sample = sample
        self._array: Optional[np.ndarray] = None
        self._actual_sample: Optional[RealNumber] = None

    def _reset_property(self) -> None:
        self._array = None
        self._actual_sample = None

    @property
    def array(self) -> np.ndarray:
        '''Representation of array axis into np.ndarray

        Returns:
            np.ndarray: numpy array
        '''
        if self._array is None:
            self._array = np.linspace(
                self._start, self._end,
                round(abs((self._end - self._start) / self._sample)) + 1
            )

        return self._array

    @property
    def start(self) -> RealNumber:
        return self._start

    @start.setter
    def start(self, value: RealNumber) -> None:
        self._start = value
        self._reset_property()

    @property
    def end(self) -> RealNumber:
        return self._end

    @end.setter
    def end(self, value: RealNumber) -> None:
        self._end = value
        self._reset_property()

    @property
    def actual_sample(self) -> RealNumber:
        if self._actual_sample is None:
            self._actual_sample = get_actual_sample(self.array)
        return self._actual_sample

    @property
    def sample(self) -> RealNumber:
        return self._sample

    @sample.setter
    def sample(self, value: RealNumber) -> None:
        self._sample = value
        self._reset_property()

    @property
    def size(self) -> int:
        return self.array.size

    def copy(self) -> 'ArrayAxis':
        '''Copy of array axis.

        Returns:
            ArrayAxis: copy of array axis
        '''
        return copy(self)

    def __str__(self):
        result = f"start: {self._start}\n" \
            f"end: {self._end}\n" \
            f"sample: {self._sample}\n" \
            f"size: {self.size}\n" \
            f"calculated sample: {self.actual_sample}"

        return result


def get_actual_sample(x: np.ndarray) -> RealNumber:
    '''Calculate actual sample. Use for self-test. Problem with floating point
    in Python (https://docs.python.org/3/tutorial/floatingpoint.html).

    Args:
        x (np.ndarray): array of numbers.

    Returns:
        RealNumber: an actual sample
    '''
    diff = np.min(np.diff(x))
    values, counts = np.unique(diff, return_counts=True)
    common_sample = values[np.argmax(counts)]
    return common_sample


def get_array_axis_from_array(
        x: ArrayLike, round_dx: bool = True) -> ArrayAxis:
    '''Create instance of Axis from some array of numbers.

    Args:
        x (ArrayLike): input array_like of numbers.
        round_dx (bool, optional): if True then round sample. Defaults to True.

    Returns:
        ArrayAxis: new ArrayAxis.
    '''

    x = np.array(x)
    dx = get_actual_sample(x)
    if round_dx and isinstance(dx, (int, float)):
        dx = 1 / round(1 / dx) if dx < 1 else round(dx)

    return ArrayAxis(start=x[0], end=x[-1], sample=dx)



import logging
from math import sqrt
from typing import Any, Callable, Optional, Union

import numpy as np

from .axis import ArrayAxis
from .config.base_config import Config
from .config.sweep_config import SweepConfig
from .defaults.sweep_methods import (CallFtatMethod, Ftatr, InterpolateArray,
                                     get_info_from_a_prior_data,
                                     get_info_from_ftat)
from .exc import BadInputError
from .relation import Relation
from .sweep import Sweep
from .help_types import ArrayLike


class UncalculatedSweep:
    '''The `UncalculatedSweep` class prepares for the calculation of the
    signal sweep (`Sweep`).

    The get_info_from_ftat function is used to extract the frequency versus
    `time` and amplitude versus `time` functions from the passed `frequency_time`
    and `amplitude_time` parameters.

    Analytic functions (`frequency_time` and `amplitude_time`) can be as array
    of numbers or as callable object(lambda function, common python functions and ect.)
    Example: frequency_time = lambda t: t*10+1, amplitude_time = lambda t: np.ones(t.size)

    Raises:
        BadInputError: raise exception when calling instance without parameter
        time or when time attribute is not created when instance initialized.
    '''

    def __init__(
        self,
        time: Union[ArrayAxis, ArrayLike] = None,
        frequency_time: Union[Ftatr, ArrayLike, InterpolateArray] = None,
        amplitude_time: Union[Ftatr, ArrayLike, InterpolateArray] = None,
    ) -> None:
        '''Initialize instance of UncalculatedSweep.

        Args:
            time (ArrayLike, optional): The `ArrayAxis`, or  an ArrayLike object
                containing numbers(real or complex). Defaults to None.

            frequency_time (Union[Ftatr, ArrayLike], optional):
                This parameter, which describes changes in frequency over time,
                can be either an array_like, or an object from which an instance
                of the `Relation` class will be created, or an instance of the
                `Relation` class, or a callable object that returns a numeric sequence.
                If `None`, then the linear function f = t will be used. Defaults to None.

            amplitude_time (Union[Ftatr, ArrayLike], optional):
                This parameter, which describes changes in amplitude modulation
                over time, can be either an array_like, or an object from which
                an instance of the `Relation` class will be created, or an instance
                of the `Relation` class, or a callable object that
                returns a numeric sequence.
                If `None`, then the function will be assumed to be constant
                and equal to 1. Defaults to None.
        '''

        self._integrate_function_default = Config.integrate_function_method
        self._get_array_axis_from_array_method = Config.get_array_axis_from_array_method

        if not (isinstance(time, ArrayAxis) or time is None):
            time = self._get_array_axis_from_array_method(time)

        if not (
            isinstance(
                frequency_time,
                (np.ndarray,
                 Relation,
                 InterpolateArray))
            or callable(frequency_time)
            or frequency_time is None
        ):
            frequency_time = np.array(frequency_time)
        if not (
            isinstance(
                amplitude_time,
                (np.ndarray,
                 Relation,
                 InterpolateArray))
            or callable(amplitude_time)
            or amplitude_time is None
        ):
            amplitude_time = np.array(amplitude_time)

        (
            self._time,
            self._frequency_time,
            self._amplitude_time,
        ) = get_info_from_ftat(time, frequency_time, amplitude_time)

    def __call__(self,
                 time: Union[ArrayAxis, ArrayLike] = None,
                 tht0=0.0
                 ) -> Sweep:
        '''Calling an instance with appropriate to calculate the sweep signal.

        If time is not passed or equals None, then the time sequence created
        when the class instance was initialized will be used.

        Args:
            time (Union[ArrayAxis, ArrayLike], optional): The number sequence
                determines the time. Defaults to None.

            tht0 (float, optional): Zero phase. Defaults to 0.0.

        Raises:
            BadInputError: raise exception when calling instance without parameter
                time when time attribute is not created when instance initialized.

        Returns:
            Sweep: an instance of the Sweep class - the calculated sweep signal.
        '''

        logging.info(
            "Calling uncalculated sweep.\n"
            "with params:\nfrequency_time={0}\namplitude_time={1}\ntime={2}"
            "".format(self._frequency_time, self._amplitude_time, time)
        )

        if time is None and self._time is None:
            raise BadInputError("Not enough data: time")

        elif time is not None:
            if isinstance(time, ArrayAxis):
                calc_time = time
            else:
                calc_time = self._get_array_axis_from_array_method(time)
        elif time is None and self._time is not None:
            calc_time = self._time

        if isinstance(self._frequency_time, InterpolateArray):
            tht = self._array_tht(self._frequency_time(calc_time))
            frequency_time = Relation(
                calc_time, self._frequency_time(calc_time))
        else:
            tht = self._func_tht(self._frequency_time)
            frequency_time = Relation(
                calc_time, self._frequency_time(
                    calc_time.array))

        if isinstance(self._amplitude_time, InterpolateArray):
            amplitude = self._amplitude_time(calc_time)
        else:
            amplitude = self._amplitude_time(calc_time.array)

        sweep = amplitude * \
            np.sin(tht(calc_time).y + tht0)

        amplitude_time = Relation(
            calc_time, amplitude
        )

        return Sweep(
            time=calc_time,
            amplitude=sweep,
            frequency_time=frequency_time,
            amplitude_time=amplitude_time,
        )

    def _func_tht(
        self, frequency_time: Callable[[np.ndarray], np.ndarray]
    ) -> Callable[[ArrayAxis], Relation]:
        """Functional representation of angular sweep."""

        def result(time: ArrayAxis) -> Relation:
            return Relation(
                *self._integrate_function_default(frequency_time, time))
        return result

    def _array_tht(
        self, frequency_time: np.ndarray
    ) -> Callable[[ArrayAxis], Relation]:
        """Angular sweep represented by a numerical sequence."""

        def wrapper(time: ArrayAxis) -> Relation:
            frequency_time_relation = Relation(time, frequency_time)
            result = np.append([0.0], 2 * np.pi *
                               frequency_time_relation.integrate().y)
            return Relation(time, result)

        return wrapper


class ApriorUncalculatedSweep(UncalculatedSweep):
    '''`ApriorUncalculatedSweep`

    Class for constructing a sweep signal from a priori data (from another
    signal or spectrum).

    The calculation of the change in frequency with time (`frequency_time`)
    and the amplitude envelope with time (`amplitude_time`) will depend on
    the a priori data (`aprior_data`) and on the method (`ftat_method`)
    by which they will be calculated.

    The extracted frequency over time (`frequency_time`) and the amplitude
    envelope over time (`amplitude_time`) will be send to the
    `UncalculatedSweep` constructor.
    '''

    def __init__(
        self,
        time: Any = None,
        a_prior_data: Any = None,
        ftat_method: Optional[CallFtatMethod] = None,
    ) -> None:
        '''Configuring an instance to create a sweep from a aprior data.

        Args:
            time (Any, optional): The number sequence
                determines the time. Defaults to None.

            a_prior_data (Any, optional): Data from which will be extracted
                frequency and amplitude modulation. Defaults to None.

            ftat_method (CallFtatMethod, optional): If the conversion method
                (`ftat_method`) is not defined or `None`, then the method is
                taken from the `SweepConfig` class `freq2time`  Method can be
                overridden if necessary
                (`sweep_design.math_signals.config.SweepConfig.freq2time`).
                Defaults to None.
        '''
        if ftat_method is None:
            ftat_method = SweepConfig.freq2time

        (
            frequency_time,
            amplitude_time,
            self._a_prior_signal,
        ) = get_info_from_a_prior_data(time, a_prior_data, ftat_method)
        super().__init__(time, frequency_time, amplitude_time)

    def __call__(
        self,
        time: Union[ArrayAxis, ArrayLike] = None,
        tht0=0.0,
        is_normalize=True
    ) -> Sweep:
        '''Calculate the sweep and normalize it.

        Args:
            time (Union[ArrayAxis, ArrayLike], optional): The number sequence
               determines the time. Defaults to None.

            tht0 (float, optional): Zero phase. Defaults to 0.0.

            is_normalize (bool, optional): Use normalization a prior data for
                sweep signal. Defaults to True.

        Returns:
            Sweep: an instance of the Sweep class - the calculated sweep signal
                from a prior data.
        '''

        sweep = super().__call__(time=time, tht0=tht0)

        if is_normalize:
            norm_sweep = sweep.get_norm()
            norm_a_prior = self._a_prior_signal.get_norm()
            norm = sqrt(norm_a_prior) / sqrt(norm_sweep)
            sweep.amplitude_time *= norm
            sweep *= norm

        sweep.a_prior_signal = self._a_prior_signal
        return sweep



from typing import Optional
import numpy as np
from scipy.signal.windows import tukey  # type: ignore

from ..help_types import Literal

Location = Literal["left", "right", "both"]


def tukey_a_t(
    time: np.ndarray,
    time_tapper: Optional[float],
    location: Location = "both",
) -> np.ndarray:
    '''Calculate array envelope for signal.

    Args:
        time (np.ndarray): time

        time_tapper (float): time_tapper in time, where coefficient will be equal 1.

        location (Literal[&quot;left&quot;, &quot;right&quot;, &quot;both&quot;], optional):
            Where the correction will be applied.
            "left" is at the start.
            "right" is at the end.
            "both" is at the start and at the end.
            Defaults to "both".

    Returns:
        np.ndarray: amplitude correction for signal. Multiple signal to result
            of function.
    '''
    if time_tapper is None:
        return np.ones(time.size)

    if time_tapper <= time[int(time.size / 2)]:
        tapper = time[time <= time_tapper].size * 2 / time.size
    else:
        tapper = 1.0

    result = tukey(time.size, alpha=tapper)

    if location == "both":
        return result

    result = np.append(
        result[: int(time.size / 2)], np.ones(time.size - int(time.size / 2))
    )

    if location == "left":
        return result

    if location == "right":
        return np.flip(result)

    return np.ones(time.size)



"""This is where default methods are defined."""
from typing import TYPE_CHECKING, Callable, Tuple, Type, Union

import numpy as np
import scipy  # type: ignore
from packaging import version
from scipy.interpolate import interp1d  # type: ignore

from sweep_design import exc  # type: ignore

from ..axis import ArrayAxis, get_array_axis_from_array
from ..help_types import X, Y
from ..core import MathOperation
from ..help_types import Number
from ..exc import TypeFuncError

if version.parse(scipy.__version__) < version.parse("1.6.0"):
    from scipy.integrate import cumtrapz, quad, trapz  # type: ignore
    integration = trapz
    cumulative_integration = cumtrapz
    quad_integrate_function = quad
else:
    from scipy.integrate import (cumulative_trapezoid, quad,  # type: ignore
                                 trapezoid)
    integration = trapezoid
    cumulative_integration = cumulative_trapezoid
    quad_integrate_function = quad

XAxis = ArrayAxis
'''Array axis of `x`.'''

FrequencyAxis = ArrayAxis
'''Array axis of frequency.'''

TimeAxis = ArrayAxis
'''Array axis of time.'''

if TYPE_CHECKING:
    from ..relation import Relation


def math_operation(
    y1: np.ndarray,
    y2: Union[np.ndarray, Number],
    name_operation: MathOperation,
) -> Y:
    '''Math operations.

    Using numpy math operations.

    Args:
        y1 (np.ndarray): first sequence y.
        y2 (Union[np.ndarray, Number]): second sequence y or other number
            name_operation (MathOperation): which mathematical operation (+, -,
            \\*, / and etc.)

    Raises:
        TypeFuncError: if operation can not be executed.

    Returns:
        Y: result of math operation.
    '''
    try:
        y = y1.__getattribute__(name_operation.value)(y2)

    except Exception as e:
        raise TypeFuncError(
            name_operation.value.strip("_"), type(y1), type(y2)) from e

    return y


def one_integrate(relation: 'Relation') -> float:
    '''Integration.

    Taking the integral on a segment. Return of the area under the graph.
    using scipy trapezoid integration.

    Args:
        relation (Relation): from will be calculated integral.

    Returns:
        float: result of integration.
    '''
    x, y = relation.get_data()
    return integration(y, x)


def integrate(relation: 'Relation') -> Tuple[XAxis, Y]:
    '''Integration.

    Integration across the entire function. Get the expected integrated
    array function.
    Using the `scipy.integrate.cumtrapz` function.

    Args:
        relation (Relation): integrated function.

    Returns:
        Tuple[XAxis, Y]: result of integration of function.
    '''
    array_axis = relation.x.copy()
    dx = array_axis.sample
    array_axis.start = array_axis.start + array_axis.sample
    return array_axis, cumulative_integration(relation.y) * (dx)


def differentiate(relation: 'Relation') -> Tuple[XAxis, Y]:
    '''Differentiation.

    The method by which differentiation is performed.
    Using the `numpy.diff` function.

    Args:
        relation (Relation): function which will be differentiated.

    Returns:
        Tuple[XAxis, Y]: result of differentiation.
    '''
    array_axis = relation.x.copy()
    dx = array_axis.sample
    array_axis.start = array_axis.start + array_axis.sample / 2
    array_axis.end = array_axis.end - array_axis.sample / 2
    return array_axis, np.diff(relation.y) / (dx)


def interpolate_extrapolate(
    x: X, y: Y, bounds_error=False, fill_value=0.0
) -> Callable[[XAxis], Y]:
    '''Interpolation and extrapolation

    Using the `scipy.interpolate.interp1d` function.
    Returning function of interpolation.

    Args:
        x (X): numbers array of axis. Samples can be not equal.

        y (Y): Representation interpolated extrapolated functions
            as array.

        bounds_error (bool, optional): if False then do not raise error if new
            array behind of bound old array. Defaults to False.

        fill_value (float, optional): default fill value if other not expected.
            Defaults to 0.0.

    Returns:
        Callable[[X], Y]: Callable that get first new array of x and return
            interpolate-extrapolate result.
    '''
    interpolate_extrapolate = interp1d(x, y, bounds_error=bounds_error,
                                       fill_value=fill_value)

    def wrapper(new_x: XAxis) -> Y:
        new_y = interpolate_extrapolate(new_x.array)
        return new_y

    return wrapper


def get_common_x(x1: XAxis, x2: XAxis) -> XAxis:
    '''Specifies the overall x-axis.

    Finds the general sample rate and beginning and end of sequence.
    A method by which to find the common sequence of numbers along
    the x-axis, obtained from two other sequences along the x-axis.

    Args:
        x1 (XAxis): first axis.
        x2 (XAxis): second axis.

    Returns:
        XAxis: return common axis.
    '''
    dx1 = x1.sample
    dx2 = x2.sample

    dx = dx1 if dx1 <= dx2 else dx2
    x_start = x1.start if x1.start <= x2.start else x2.start
    x_end = x1.end if x1.end >= x2.end else x2.end
    return ArrayAxis(start=x_start, end=x_end, sample=dx)


def correlate(cls: Type["Relation"], r1: "Relation",
              r2: "Relation") -> Tuple[XAxis, np.ndarray]:
    '''Correlation.

    The method by which the correlation is performed.
    Using the `numpy.correlate` function.

    Args:
        cls (Type[&quot;Relation&quot;]): class to use equalization of two arrays.
        r1 (Relation): first function y.
        r2 (Relation): second function y.

    Returns:
        Tuple[XAxis, np.ndarray]: result of correlation.
    '''
    r1 = r1.shift(-r1.start)
    r2 = r2.shift(-r2.start)
    r1, r2 = cls.equalize(r1, r2)
    x_axis = ArrayAxis(start=-r1.end, end=r1.end, sample=r1.sample)
    return x_axis, np.correlate(r1.y, r2.y, "full")


def convolve(cls: Type["Relation"], r1: "Relation",
             r2: "Relation") -> Tuple[XAxis, np.ndarray]:
    '''Convolution.

    The method by which the convolution is performed.
    Using the `numpy.convolve` function.

    Args:
        cls (Type[&quot;Relation&quot;]): class to use equalization of two arrays.
        r1 (Relation): first function y.
        r2 (Relation): second function y.

    Returns:
        Tuple[XAxis, np.ndarray]: result of convolution.
    '''
    r1 = r1.shift(-r1.start)
    r2 = r2.shift(-r2.start)
    r1, r2 = cls.equalize(r1, r2)
    x_axis = ArrayAxis(start=-r1.end, end=r1.end, sample=r1.sample)
    return x_axis, np.convolve(r1.y, r2.y, "full")


# ==============================================================================

def _calculate_spectrum(
        time: TimeAxis, amplitude: np.ndarray) -> Tuple[FrequencyAxis, np.ndarray]:

    amplitude = np.append(
        amplitude[time.array >= 0.0], amplitude[time.array < 0.0])
    spectrum = np.fft.rfft(amplitude)
    np_frequency = np.fft.rfftfreq(
        amplitude.size, d=time.sample
    )

    frequency = get_array_axis_from_array(np_frequency, round_dx=False)
    return frequency, spectrum


def signal2spectrum(
    relation: 'Relation', is_start_zero=False
) -> Tuple[FrequencyAxis, np.ndarray]:
    '''Forward Fourier Transform.

    Method for converting a signal into a spectrum.
    Using the `numpy.fft.rfft` function.

    Args:
        relation (Relation): signal from which get spectrum.
        is_start_zero (bool, optional): Consider array started from zero time.
            Defaults to False.

    Returns:
        Tuple[FrequencyAxis, np.ndarray]: result transformation signal to
            spectrum.
    '''
    new_time = relation.x.copy()
    amplitude = relation.y.copy()
    if is_start_zero:
        return _calculate_spectrum(new_time, amplitude)

    if new_time.start > 0.0:
        new_time.start = 0.0
        amplitude = np.append(
            np.zeros(new_time.size - amplitude.size),
            amplitude)

    elif new_time.end < 0.0:

        new_time.end = 0.0
        amplitude = np.append(
            amplitude, np.zeros(
                new_time.size - amplitude.size))

    return _calculate_spectrum(new_time, amplitude)


def spectrum2signal(
    relation: 'Relation', time_start: float = None
) -> Tuple[TimeAxis, np.ndarray]:
    '''Inverse Fourier Transform.

    Method for converting a spectrum into a signal.
    Using `numpy.ifft` function.

    Args:
        relation (Relation): spectrum of signal.

        time_start (float, optional): default fft convert to 0. time. Maybe you
            want another start of time. Defaults to None.

    Returns:
        Tuple[TimeAxis, np.ndarray]: result transformation spectrogram to signal.
    '''

    spectrum = relation.y.copy()
    frequency = relation.x.copy()

    amplitude = np.fft.irfft(spectrum)  # type: np.ndarray
    dt = 1 / (2 * (frequency.end - frequency.start))

    if time_start is None:
        time = ArrayAxis(0., (amplitude.size - 1) * dt, dt)

    else:
        time = ArrayAxis(time_start, time_start +
                         (amplitude.size - 1) * dt, dt)

        amplitude = np.append(
            amplitude[time.array >= 0.0], amplitude[time.array < 0.0])

    return time, amplitude


def integrate_function(
    function: Callable[[np.ndarray], np.ndarray], x: ArrayAxis
) -> Tuple[ArrayAxis, np.ndarray]:
    '''Integration function y(x).

    The method by which the integration function is performed. Integration across
    the entire function. Get the expected integrated array function.
    Integration of function, using `scipy.integrate.quad` function.

    Args:
        function (Callable[[x], y]): function is describing
            changes frequency from time.

        x (np.ndarray): time array.

    Returns:
        Relation: result of integration function.
    '''

    result = np.append(
        [0.0],
        np.array(
            [2 * np.pi * quad_integrate_function(function, x.start, x_element)[0]
             for x_element in x.array[1:]]
        ),
    )
    return x, result



from .dwell_sweep import get_dwell_sweep
from .linear_sweep import get_linear_sweep
from .pseudorandom_shuffle import get_shuffle
from .sweep_from_code import get_code_sweep_segments, get_convolution_sweep_and_code
from .code_m_sequence import get_m_sequence_code, get_relation_m_sequence
from .code_zinger import get_code_zinger, get_code_zinger_relation



from typing import Any, Callable, Optional, Tuple, Union

import numpy as np
from scipy.signal import hilbert, spectrogram  # type: ignore

from .. import relation, signal, spectrum
from ..axis import ArrayAxis
from ..config import base_config
from ..exc import BadInputError
from ..help_types import ArrayLike, Frequency, Time, Envelope, Spectrogram, Ftat

Ftatr = Union["relation.Relation", Ftat]
'''The Representation of object from which will be extracted frequency
modulation or amplitude modulation.'''

CallFtatMethod = Callable[[
    "spectrum.Spectrum"], Tuple[Time, Frequency, Envelope]]
'''Method is used to extract frequency and amplitude modulation from spectrum.'''


def simple_freq2time(
        spectrum: "spectrum.Spectrum") -> Tuple[Time, Frequency, Envelope]:
    '''The simple method to extract Frequency modulation from a prior spectrum.

    The amplitude modulation is constant.

    Args:
        spectrum (Spectrum): instance signal of 'Spectrum'

    Returns:
        Tuple[Time, Frequency, Envelope]: simple representation Frequency
            modulation from a prior spectrum.
    '''

    amplitude_spectrum = spectrum.get_amp_spectrum()
    frequency = amplitude_spectrum.array
    n_spec = amplitude_spectrum.y**2
    time = np.append(
        [0.0], ((n_spec[1:] + n_spec[:-1]) /
                (frequency[1:] - frequency[:-1])).cumsum()
    )
    coefficient = amplitude_spectrum.get_norm()
    amplitude_modulation = coefficient * np.ones(len(time))
    return time, frequency, amplitude_modulation


class InterpolateArray:
    '''Class to interpolate array.

    Use to stretch data to new axis.
    '''

    def __init__(
        self,
        x: Union[relation.Relation, ArrayAxis, ArrayLike],
        y: Optional[ArrayLike] = None
    ) -> None:
        '''Initialize InterpolateArray.

        Args:
            x (Union[relation.Relation, ArrayAxis, ArrayLike]): old array of x
                or old `Relation` instance.

            y (Optional[ArrayLike], optional): old array of y or None. Defaults to None.

        Raises:
            BadInputError: raise exception if will be not enough data.
        '''

        if isinstance(x, relation.Relation):
            self._x = x.x.array.copy()
            self._y = x.y.copy()
            return None

        elif y is None:
            raise BadInputError("Not enough data! y is absent")

        self._y = np.array(y)

        if isinstance(x, ArrayAxis):
            self._x = x.array.copy()
        else:
            self._x = np.array(x)

    def __call__(self, new_x: ArrayAxis) -> np.ndarray:
        '''Get new array of y.

        Stretch and interpolate old array y to get new array y.
        Args:
            new_x (ArrayAxis): new array of x.

        Returns:
            np.ndarray: new array of y.
        '''
        stretch_old_x = self._x * \
            ((new_x.end - new_x.start) /
             self._x[-1]) - self._x[0] + new_x.start

        new_y = base_config.Config. \
            interpolate_extrapolate_method(stretch_old_x, self._y)

        return new_y(new_x)


def convert_freq2time(
    spectrum: "spectrum.Spectrum", convert_method: CallFtatMethod
) -> Tuple[InterpolateArray, InterpolateArray]:
    '''Convert spectrum to function to create sweep.

        The method consists in obtaining the functions of Frequency and
        amplitude modulation.

    Args:
        spectrum (Spectrum): spectrum a prior data.

        convert_method (CallFtatMethod): method to convert spectrum into
            frequency-time and amplitude-time arrays.

    Returns:
        TupleInterpolateArray,InterpolateArray]:
            tuple of two function. First function describe changes frequency
            from time. Second function describe changes amplitude from time.
            This function will be use to create sweep signal.

    '''
    nT, f, a_t = convert_method(spectrum)
    return InterpolateArray(nT, f), InterpolateArray(nT, a_t)


def get_info_from_a_prior_data(
    time: Union["relation.Relation", ArrayAxis, ArrayLike],
    a_prior_data: Any,
    f_a_t_method: CallFtatMethod
) -> Tuple[InterpolateArray, InterpolateArray, "signal.Signal"]:
    '''Get the Frequency and amplitude modulation function from a prior data.

    Args:
        time (Any): time.

        a_prior_data (Any): a prior data.

        f_a_t_method (CallFtatMethod): method to convert a prior data for
            functions that will use to create sweep signal.

    Returns (TupleInterpolateArray,InterpolateArray, "Signal"]):
        return frequency-time and amplitude-time functions and a prior signal.
    '''

    if isinstance(time, spectrum.Spectrum):
        a_prior_signal = time.get_signal()
        a_prior_spectrum = time
    elif isinstance(time, relation.Relation):
        a_prior_signal = signal.Signal(time)
        a_prior_spectrum = a_prior_signal.get_spectrum()
    elif isinstance(a_prior_data, spectrum.Spectrum):
        a_prior_signal = a_prior_data.get_signal()
        a_prior_spectrum = a_prior_data
    elif isinstance(a_prior_data, signal.Signal):
        a_prior_signal = a_prior_data
        a_prior_spectrum = a_prior_data.get_spectrum()
    elif isinstance(a_prior_data, relation.Relation):
        a_prior_spectrum = spectrum.Spectrum(a_prior_data)
        a_prior_signal = a_prior_spectrum.get_signal()
    else:
        a_prior_signal = signal.Signal(time, a_prior_data)
        a_prior_spectrum = a_prior_signal.get_spectrum()

    f_t, a_t = convert_freq2time(a_prior_spectrum, f_a_t_method)

    return f_t, a_t, a_prior_signal


def get_info_from_ftat(
    t: Union[ArrayAxis, np.ndarray, None] = None,
    f_t: Union[Ftatr, InterpolateArray, None] = None,
    a_t: Union[Ftatr, InterpolateArray, None] = None
) -> Tuple[Optional[ArrayAxis],
           Union[Callable[[np.ndarray], np.ndarray], InterpolateArray],
           Union[Callable[[np.ndarray], np.ndarray], InterpolateArray]
           ]:
    '''Prepared information for 'UncalculatedSweep' instance.

    Args:
        t (Optional[np.ndarray]): time. Default None.
        f_t (Optional[Ftatr]): frequency-time function. Default None.
        a_t (Optional[Ftatr]): amplitude-time function. Default None.

    Returns:
        Tuple[Optional[np.ndarray],InterpolateArray,InterpolateArray]:
        return frequency-time and amplitude-time functions and a prior signal.
    '''
    if f_t is None:

        def linear_time(time: np.ndarray) -> np.ndarray:
            return time

        f_t = linear_time

    t1, f_t = _extract_x_t(t, f_t)

    if a_t is None:

        def const_one(time: np.ndarray) -> np.ndarray:
            return np.ones(len(time))

        a_t = const_one

    t2, a_t = _extract_x_t(t, a_t)

    if isinstance(t, np.ndarray):
        t = base_config.Config.get_array_axis_from_array_method(t, True)
    if t is None:
        if t1 is None and t2 is not None:
            t = t2
        elif t2 is None and t1 is not None:
            t = t1
        elif t2 is not None and t1 is not None:
            t = base_config.Config.get_common_x(t1, t2)

    return t, f_t, a_t


def _extract_x_t(
    t: Union[ArrayAxis, np.ndarray, None] = None,
    x_t: Union[Ftatr, InterpolateArray] = None,
) -> Tuple[
    Optional[ArrayAxis],
    Union[Callable[[np.ndarray], np.ndarray], InterpolateArray]
]:
    if isinstance(t, np.ndarray):
        t = base_config.Config.get_array_axis_from_array_method(t, False)

    b_x_t: Union[Callable[[np.ndarray], np.ndarray], InterpolateArray]
    if not callable(x_t):
        if isinstance(x_t, InterpolateArray):
            b_x_t = InterpolateArray
        elif isinstance(x_t, relation.Relation):
            b_x_t = InterpolateArray(x_t)
        else:
            if isinstance(t, np.ndarray):
                t = base_config.Config.get_array_axis_from_array_method(t)

            if isinstance(t, ArrayAxis) and isinstance(x_t, np.ndarray):
                if t.size != x_t.size:
                    calc_t = ArrayAxis(
                        t.start, t.end, (t.start - t.end) / (x_t.size - 1))

                    interpolate_x_t = base_config.Config. \
                        interpolate_extrapolate_method(calc_t.array, x_t)
                    x_t = interpolate_x_t(t)
                b_x_t = InterpolateArray(t, x_t)
            else:
                raise BadInputError("Not enough data: t or x_t")
    else:
        b_x_t = x_t

    return t, b_x_t

# ======================================================================
# Info about sweep.


def get_spectrogram(sweep: "relation.Relation") -> Spectrogram:
    '''Function to get spectrogram of the sweep signal.

    Using the `scipy.signal.spectrogram` function.

    Args:
        sweep (relation.Relation): instance of sweep signal.

    Returns:
        Spectrogram: tuple of np.ndarray. The first element is time.
            The second is frequency. The third is matrix M x N of spectrogram.
    '''

    if sweep.y.size < 256:
        nperseg = sweep.y.size
    else:
        nperseg = 256
    frequency, spectrogram_time, spectrogram_ = spectrogram(
        sweep.y, 1 / (sweep.sample), nperseg=nperseg)

    return spectrogram_time, frequency, spectrogram_[::-1, ::]


def get_f_t(sweep: "relation.Relation") -> "relation.Relation":
    '''Get Time-Frequency function from sweep signal using the Hilbert
    transformation.

    Using the `scipy.signal.hilbert` function.

    Args:
        sweep (Relation): instance of sweep signal.

    Returns:
        Relation: instance `Relation`
    '''

    analytical_signal = hilbert(sweep.y)
    result = np.append(
        [0.0],
        np.diff(np.unwrap(np.angle(analytical_signal)))
        / (2.0 * np.pi)
        / sweep.sample,
    )
    return relation.Relation(sweep.x, result)


def get_a_t(sweep: "relation.Relation") -> "relation.Relation":
    '''Get envelop from sweep signal using the Hilbert transformation.

    Using the `scipy.signal.hilbert` function.

    Args:
        sweep (Relation): instance of sweep signal.

    Returns:
        Relation: instance `Relation`
    '''

    analytical_signal = hilbert(sweep.y)
    return relation.Relation(sweep.x, np.abs(analytical_signal))



from ..signal import Signal
from .emd_analyze import get_IMFs_emd
from .a_t import tukey_a_t


def correct_sweep(signal: Signal, start_window: float = None) -> Signal:
    '''Apply correction to sweep signal.

    Using the EMD to subtract the last IMF from the displacement and
    if window is not None then apply a window in the star so that the
    displacement starts at zero.

    Args:
        signal (Relation): The sweep signal on which the correction will be
            applied.

        start_window (float, optional): The time interval at the beginning to
            reduce the deviation from zero. If None, then window is not applied.
            Defaults to None.

    Returns:
        Relation: corrected sweep signal.
    '''

    displacement = signal.integrate().integrate()
    x = displacement.x.array

    IMFs = get_IMFs_emd(displacement)
    if start_window is not None:
        window = tukey_a_t(x, start_window, "left")
        new_displacement = IMFs[0] * Signal(x, window)
    else:
        new_displacement = IMFs[0]

    signal = new_displacement.diff().diff()

    return signal



class NotEqualError(Exception):
    def __init__(self, *args: object) -> None:
        message = f"Different size x_array({args[0]}) and y_array({args[1]})"
        super().__init__(message)


class BadInputError(Exception):
    pass


class ConvertingError(Exception):
    def __init__(self, *args: object) -> None:
        message = f"Can not convert type {args[0]} into type {args[1]}."
        super().__init__(message)


class TypeFuncError(Exception):
    def __init__(self, *args: object) -> None:
        message = f'operation "{args[0]}" did not complete with type'
        '{args[1]} and type {args[2]}'
        super().__init__(message)



from math import sqrt
from typing import Callable, Tuple

import numpy as np
from scipy import integrate  # type: ignore

from ..spectrum import Spectrum
from ..help_types import Time, Frequency, Envelope


def proportional_freq2time(
        spectrum: "Spectrum") -> Tuple[Time, Frequency, Envelope]:
    """Convert spectrum to frequency-time and amplitude-time arrays.

    A function to get frequency versus time and amplitude versus time changes
    from an a priori spectrum to create a sweep signal. **In this implementation,
    the amplitude of change over time is a constant.** The spectrum of the
    resulting sweep signal will be equal to the prior spectrum.
    """
    f, a_f = spectrum.get_amp_spectrum().get_data()
    n_spec = a_f**2
    n_t = np.append(
        [0.0], ((n_spec[1:] + n_spec[:-1]) / (f[1:] - f[:-1])).cumsum())
    coefficient = integrate.trapz(n_spec, f)
    a_t = sqrt(coefficient * 2) * np.ones(len(n_t))
    return n_t, f, a_t


def dwell(f_start: float = None, f_end: float = None, fc: float = None) \
        -> Callable[[Spectrum], Tuple[Time, Frequency, Envelope]]:
    """Convert spectrum to frequency-time and amplitude-time arrays.

    A function to get frequency versus time and amplitude versus time changes
    from an a priori spectrum to create a sweep signal. **In this
    implementation, the amplitude of the envelope monotonously increases
    in proportion to the change in frequency up to the cutoff frequency fc,
    after which the amplitude of the envelope is constant.** The spectrum of the
    resulting sweep signal will be equal to the prior spectrum.
    """

    def freq2time(
            spectrum: "Spectrum") -> Tuple[Time, Frequency, Envelope]:

        f, a_f = spectrum.select_data(
            f_start, f_end).get_amp_spectrum().get_data()

        f_c = fc if fc else f[0]

        freq = f[f <= f_c]

        freq_ratio = np.divide(
            freq[1:], freq[:-1], out=np.ones_like(freq[1:]), where=freq[:-1] != 0
        )
        dur = np.cumprod(1 / (freq_ratio**4))
        energy = a_f[f <= f_c][:-1] ** 2
        nT = np.append([0.0], dur * energy).cumsum()
        nT2 = np.append(nT[-1], dur[-1] * (a_f[f > f_c] ** 2)).cumsum()
        nT = np.append(nT, nT2[1:])

        a_t = np.cumprod(np.append([1.0], freq_ratio**2))
        a_t = np.append(a_t, [a_t[-1]] * (len(a_f) - len(a_t)))

        return nT, f, a_t

    return freq2time




from typing import NamedTuple

import numpy as np

from .axis import ArrayAxis


class Spectrogram(NamedTuple):
    '''Spectrogram of signal.

    Gathering time, frequency and spectrogram image together.
    '''
    time: ArrayAxis
    frequency: ArrayAxis
    spectrogram: np.ndarray



from typing import List, Union

import numpy as np

from sweep_design.exc import BadInputError

from ..help_types import Literal
from ..axis import ArrayAxis
from ..relation import Relation
from ..sweep import Sweep


def get_convolution_sweep_and_code(
    code: Union[Relation, List[Literal[-1, 0, 1]]],
    base_sweep: Relation
) -> Sweep:
    '''Get sweep use convolution between sweep and desired code.

    Code can be m-sequence, code zinger and etc.

    Args:
        code (Union[Relation, List[Literal[-1, 0, 1]]): some code.
            Relation or List of -1, 0, 1

        base_sweep (Relation): some base sweep.

    Returns:
        Sweep: result signal use as sweep signal.
    '''
    if not isinstance(code, Relation):
        time_code = ArrayAxis(
            0,
            (len(code) - 1) *
            base_sweep.sample,
            base_sweep.sample)
        code = Relation(time_code, code)

    result = Sweep.convolve(
        code,
        base_sweep)[None:code.size *
                    base_sweep.sample]  # type: ignore

    if isinstance(result, Sweep):
        return result.shift(
            base_sweep.end)

    raise BadInputError('Not enough data!')


def get_code_sweep_segments(
    code: Union[Relation, np.ndarray, List[Literal[-1, 0, 1]]],
    base_sweep: Relation
) -> Sweep:
    '''Get sweep use code to compose a base_sweep.

    Code can be m-sequence, code zinger and etc.
    Args:
        code (Union[Relation, List[Literal[-1, 0, 1]]): some code.
        sweep (Relation): some base sweep.
    '''
    if isinstance(code, Relation):
        code = code.y

    new_sweep = base_sweep * code[0]

    for cnt, v in enumerate(code[1:], 1):

        new_sweep = new_sweep + base_sweep.shift(
            cnt * base_sweep.end + base_sweep.sample
        ) * v

    return Sweep(new_sweep)



import logging
from copy import deepcopy
from typing import Tuple, Type, TypeVar, Union

import numpy as np

from .axis import ArrayAxis
from .config.base_config import Config
from .core import MathOperation, RelationProtocol
from .exc import BadInputError, NotEqualError, TypeFuncError
from .help_types import ArrayLike, Number, RealNumber

R = TypeVar("R", bound="Relation")
'''Description first `Relation`.'''

R2 = TypeVar("R2", bound="Relation")
'''Description second `Relation`.'''


class Relation(RelationProtocol):
    '''A representation of dependency y from x (y = f(x)).

    The class describe the dependency between x, y. x is `ArrayAxis` instance showing
    the start of sequence, the end of sequence and the sample space between elements.
    They consist of real or complex numbers. The array length calculated from
    `ArrayAxis` must be equal length of y sequence.

    For the instance of `Relation` class, define the basic mathematical operations:
    addition (+), subtraction(-), multiplication(\\*), division(/),
    exponentiation (\\*\\*) and their unary representation (+=, -=, \\*=, /=).
    The result of the operation is a new instance of the `Relation` class.

    Determined correlation and convolution between two instances
    (methods: correlate and convolve).

    How those operations will be calculated determined by the methods described
    in the Config class. Methods can be overridden if necessary
    (sweep-design.config).

    WARNING!!! When inheriting the `Relation` class, it is important to write correctly
    constructor. It must match the constructor of the `Relation` class.
    Because some methods return a type(self)(...). For example,
    addition method (def __add__(self: R, other: Union['Relation', Num]) -> R).
    Or predefine these methods in the inherited class.

    Raises:
        BadInputError: Raise this exception if we don't have enough data.
        NotEqualError: Raise this exception if we try create instance use
            different length of sequence numbers for x and y.
        TypeFuncError: Raise an exception, when execute some function with
            unexpected type of value.

    Returns:
        _type_: Type of Relation.
    '''

    def __init__(
        self,
        x: Union[RelationProtocol, ArrayAxis, ArrayLike],
        y: ArrayLike = None,
    ) -> None:
        '''Initialization of instance of `Relation`.

        Args:
            x (Union[RelationProtocol, ArrayLike, ArrayAxis]):
                The `Relation` class, or a class derived from the `Relation`
                class, or instance of `ArrayAxis` or an `ArrayLike` object
                containing numbers(real or complex). if x is `ArrayLike` then
                it will be converted to `ArrayAxis` instance use method
                *get_array_axis_from_array_method* from Config class

            y (ArrayLike, optional):
                None or array_like object containing real or complex numbers.
                If it is not None then it will be converted to np.ndarray.
                Defaults to None.

        Raises:
            BadInputError:Raise this exception if we don't have enough data.
            NotEqualError: Raise this exception if we try create instance use
        '''

        self._get_array_axis_from_array_method = Config.get_array_axis_from_array_method
        self._math_operation = Config.math_operation
        self._interpolate_extrapolate_method = Config.interpolate_extrapolate_method
        self._integrate_one_method = Config.integrate_one_method
        self._integrate_method = Config.integrate_method
        self._differentiate_method = Config.differentiate_method

        if isinstance(x, RelationProtocol):
            self._x = x.x.copy()
            self._y = x.y.copy()
            if y is not None:
                logging.warning(f'x is instance of {type(x)}, "y" was ignored')
            return None

        if y is None:
            raise BadInputError("y is absent. Not enough data!")

        y = np.array(y)

        if not isinstance(x, ArrayAxis):
            x = self._get_array_axis_from_array_method(x)

        if x.size != y.size:
            raise NotEqualError(x.size, y.size)

        self._x, self._y = x, y

    @property
    def x(self) -> ArrayAxis:
        '''ArrayAxis of relation.

        Returns:
            ArrayAxis: array axis of relation.
        '''
        return self._x

    @property
    def y(self) -> np.ndarray:
        '''Result of relation of y(x)

        Returns:
            np.ndarray: array of numbers represent relation of y(x)
        '''
        return self._y

    @property
    def start(self) -> RealNumber:
        '''Start of array axis x.

        Returns:
            RealNumber: start number of array axis x.
        '''
        return self._x.start

    @start.setter
    def start(self, value: RealNumber) -> None:
        '''Setter for start.

        Args:
            value (RealNumber): set start for array axis x.
        '''
        self._x.start = value

    @property
    def end(self) -> RealNumber:
        '''End of array axis x.

        Returns:
            RealNumber: end number of array axis x.
        '''
        return self._x.end

    @end.setter
    def end(self, value: RealNumber):
        '''Setter for start.

        Args:
            value (RealNumber): set end for array axis x.
        '''
        self._x.end = value

    @property
    def sample(self) -> RealNumber:
        '''Sample for array axis x.

        Returns:
            RealNumber: sample of array axis x.
        '''
        return self._x.sample

    @sample.setter
    def sample(self, value: RealNumber) -> None:
        '''Setter for sample.

        Args:
            value (RealNumber): set sample for array axis x.
        '''
        self._x.sample = value

    @property
    def array(self) -> np.ndarray:
        '''Get array representation of array axis x.

        Returns:
            np.ndarray: array of numpy.
        '''
        return self._x.array

    @property
    def actual_sample(self) -> Number:
        '''Get actual sample or array axis x.

        Returns:
            Number: number of actual sample array x.
        '''
        return self._x.actual_sample

    @property
    def size(self) -> int:
        '''size of array axis x.

        Returns:
            int: integer number of array size x.
        '''
        return self._x.size

    def get_data(self) -> Tuple[np.ndarray, np.ndarray]:
        '''Return the data of the object.

        Raises:
            NotEqualError: After manipulating on x ArrayAxis, the size of the
                extracted arrays is checked. If they are different then raise
                that error.

        Returns:
            Tuple[np.ndarray, np.ndarray]: tuple of two number sequence
        '''

        if self._x.size != self._y.size:
            raise NotEqualError(self._x.size, self._y.size)

        return self.array.copy(), self.y.copy()

    def max(self) -> Number:
        '''Get maximum of Relation.

        Returns:
            Number: maximum of y array.
        '''
        return self._y.max()

    def min(self) -> Number:
        '''Get minimum of Relation.

        Returns:
            Number: minimum of y array.
        '''
        return self._y.min()

    def get_norm(self) -> RealNumber:
        '''Get signal rate.

        Calculated in terms of signal energy.

        Returns:
            Number: signal rate
        '''

        square_self = self**2
        return self._integrate_one_method(square_self) / (self.sample)

    def select_data(self: R, start: Number = None,
                    end: Number = None) -> R:
        '''Select data using x-axis

        Args:
            self (R): instance of Relation
            start (Number, optional): new start of relation x. Defaults to None.
            end (Number, optional): new end of relation x. Defaults to None.

        Returns:
            R: new instance of Relation.
        '''

        if start is None:
            start = self.start

        if end is None:
            end = self.end

        array = self.array

        is_selected = np.logical_and(
            np.greater_equal(array, start), np.less_equal(array, end)
        )
        selected_x = array[is_selected]

        new_x_array = ArrayAxis(
            start=selected_x[0], end=selected_x[-1], sample=self.sample)

        return type(self)(new_x_array, self.y[is_selected])

    def exp(self: R) -> R:
        '''Get exponent of Relation.

        Args:
            self (R): instance of Relation

        Returns:
            R: Relation where new y is exponent of old y.
        '''
        return type(self)(self.x.copy(), np.exp(self.y))

    def diff(self: R) -> R:
        '''Differentiation of 'Relation'.

        Args:
            self (R): instance of Relation

        Returns:
            R: result of differentiation.
        '''
        result = self._differentiate_method(self)
        return type(self)(*result)

    def integrate(self: R) -> R:
        '''Integration of `Relation`.

        Args:
            self (R): instance of Relation

        Returns:
            R: result of cumulative integration.
        '''
        result = self._integrate_method(self)
        return type(self)(*result)

    def interpolate_extrapolate(
            self: R, new_x: Union[R, ArrayAxis, ArrayLike]) -> R:
        '''Interpolates and extrapolates an existing relation using new array
        x of the represented ArrayAxis instance.

        Args:
            self (R): instance of Relation
            new_x (ArrayAxis): new x array axis

        Returns:
            R: new instance of Relation
        '''
        if isinstance(new_x, Relation):
            new_x = new_x.x.copy()
        elif isinstance(new_x, ArrayAxis):
            new_x = new_x.copy()
        else:
            new_x = self._get_array_axis_from_array_method(new_x, False)

        new_y = self._interpolate_extrapolate_method(
            self.x.array.copy(), self.y.copy())(new_x)
        return type(self)(new_x, new_y)

    def shift(self: R, x_shift: RealNumber = 0) -> R:
        '''Shifting of relation on the x-axis.

        Args:
            self (R): instance of Relation
            x_shift (Number, optional): Number of displacement on the x-axis.
            Defaults to 0.

        Returns:
            R: new instance of Relation
        '''
        new_x = self.x.copy()
        new_x.start = new_x.start + x_shift
        new_x.end = new_x.end + x_shift
        return type(self)(new_x, self.y)

    @staticmethod
    def equalize(r1: R, r2: R2) -> Tuple[R, R2]:
        '''Bringing two Relation objects with different x-axes to one common one.

        When converting, interpolation and extrapolation are used.

        Args:
            r1 (R): first instance of Relation
            r2 (R2): second instance of Relation

        Returns:
            Tuple[R, R2]: tuple of new Relation instances with common axis.
        '''
        if r1.x == r2.x:
            return deepcopy(r1), deepcopy(r2)

        new_x = Config.get_common_x(r1.x, r2.x)
        r1 = r1.interpolate_extrapolate(new_x)
        r2 = r2.interpolate_extrapolate(new_x)

        return r1, r2

    @classmethod
    def correlate(cls: Type[R], r1: "Relation", r2: "Relation") -> R:
        '''Correlation of two Relations.

        Args:
            cls (Type[R]): class of Relation
            r1 (Relation): first Relation.
            r2 (Relation): second Relation.

        Raises:
            TypeFuncError: raise exception
            if we try correlate with unexpected types.

        Returns:
            R: new instance of Relation
        '''

        if isinstance(r1, Relation) and isinstance(r2, Relation):
            result = Config.correlate_method(cls, r1, r2)
            return cls(*result)
        else:
            raise TypeFuncError("Correlation", type(r1), type(r2))

    @classmethod
    def convolve(cls: Type[R], r1: "Relation", r2: "Relation") -> R:
        '''Convolution of two Relations.

        Args:
            cls (Type[R]): class of Relation
            r1 (Relation): first Relation.
            r2 (Relation): second Relation.

        Raises:
            TypeFuncError: raise exception
            if we try correlate with unexpected types.

        Returns:
            R: new instance of Relation
        '''
        if isinstance(r1, Relation) and isinstance(r2, Relation):
            result = Config.convolve_method(cls, r1, r2)
            return cls(*result)
        else:
            raise TypeFuncError("Convolution", type(r1), type(r2))

    @staticmethod
    def _operation(
        a: "Relation", b: Union["Relation", Number], name_operation: MathOperation
    ) -> Tuple[ArrayAxis, np.ndarray]:
        logging.debug(f"Type of a: {type(a)}")
        logging.debug(f"Type of b: {type(b)}")

        if isinstance(b, RelationProtocol):
            r1, r2 = Relation.equalize(a, b)
            return r1.x.copy(), a._math_operation(
                r1.y.copy(), r2.y.copy(), name_operation)
        else:
            return a.x.copy(), a._math_operation(
                a.y.copy(), b, name_operation)

    def __add__(self: R, other: Union["Relation", Number]) -> R:
        return type(self)(
            *self._operation(self, other, MathOperation.ADD))

    def __radd__(self: R, other: Union["Relation", Number]) -> R:
        return type(self)(*self._operation(self, other,
                                           MathOperation.RADD))

    def __sub__(self: R, other: Union["Relation", Number]) -> R:
        return type(self)(
            *self._operation(self, other, MathOperation.SUB))

    def __rsub__(self: R, other: Union["Relation", Number]) -> R:
        return type(self)(*self._operation(self, other,
                                           MathOperation.RSUB))

    def __mul__(self: R, other: Union["Relation", Number]) -> R:
        return type(self)(
            *self._operation(self, other, MathOperation.MUL))

    def __rmul__(self: R, other: Union["Relation", Number]) -> R:
        return type(self)(*self._operation(self, other,
                                           MathOperation.RMUL))

    def __truediv__(self: R, other: Union["Relation", Number]) -> R:
        return type(self)(
            *self._operation(self, other, MathOperation.TRUEDIV)
        )

    def __rtruediv__(self: R, other: Union["Relation", Number]) -> R:
        return type(self)(
            *self._operation(self, other, MathOperation.RTRUEDIV)
        )

    def __pow__(self: R, other: Union["Relation", Number]) -> R:
        return type(self)(
            *self._operation(self, other, MathOperation.POW))

    def __rpow__(self: R, other: Union["Relation", Number]) -> R:
        return type(self)(*self._operation(self, other,
                                           MathOperation.RPOW))

    def __iadd__(self: R, other: Union["Relation", Number]) -> R:
        return self.__add__(other)

    def __isub__(self: R, other: Union["Relation", Number]) -> R:
        return self.__sub__(other)

    def __imul__(self: R, other: Union["Relation", Number]) -> R:
        return self.__mul__(other)

    def __idiv__(self: R, other: Union["Relation", Number]) -> R:
        return self.__truediv__(other)

    def __ipow__(self: R, other: Union["Relation", Number]) -> R:
        return self.__pow__(other)

    def __len__(self) -> int:
        return self._x.size

    def __getitem__(
            self: R, select_data: Union[Number, slice]) -> Union[Tuple[Number, Number], R]:
        '''Select data from Relation

        if item is Number then function return tuple of two numbers.
        The first number is number near to select data.
        Second number is number represent of relation to selected data.

        if select data is slice then function return Relation that
        equal Relation if we call select_data function of instance.

        Args:
            self (R): instance of Relation
            item (Union[float, slice]): selected data is number or slice

        Returns:
            Union[Tuple[Num, Num], R]: two number or instance of relation.
        '''

        if isinstance(select_data, (float, int, complex)):
            array = self.array
            idx = (np.abs(array - select_data)).argmin()
            return array[idx], self.y[idx]

        if isinstance(select_data, slice):
            return self.select_data(select_data.start, select_data.stop)

    def __str__(self) -> str:
        return f"y: {self.y}\nx: {str(self.x)}"



from typing import List

from ..help_types import Literal

from ..relation import Relation


def get_code_zinger(
    first_sequence_code_zinger: List[Literal[-1, 1]] = [-1, -1, -1, 1],
    periods=1
) -> List[Literal[-1, 1]]:
    '''Build sequence of code zinger.

    Args:
        first_sequence_code_zinger (Literal[, optional): First period of code
            zinger. Defaults to [-1, -1, -1, 1].

        periods (int, optional): Periods of code zinger.
            Repeat first_sequence_code_zinger n(periods) times. Defaults to 1.

        Returns:
            np.ndarray: array of code zinger.
    '''
    return first_sequence_code_zinger * periods


def get_code_zinger_relation(start_sequence: Relation, periods=1) -> Relation:
    '''Create relation of code zinger.

    Args:
        start_sequence (Relation): start of code zinger.
            Example: Relation([-1, -1, -1, 1], [0, 1, 2, 3])

        periods (int, optional):  Periods of code zinger.
            Repeat first_sequence_code_zinger n(periods) times. Defaults to 1.

        Returns:
            Relation: relation of code zinger.
    '''
    sequence = start_sequence

    if periods > 1:
        for k in range(periods):
            sequence = sequence + start_sequence.shift(start_sequence.end * k)

    return sequence



from typing import List

__version__ = '0.2.0'
__all__: List[str] = []

from .axis import ArrayAxis as ArrayAxis
from .relation import Relation as Relation

from .spectrum import Spectrum as Spectrum
from .signal import Signal as Signal
from .sweep import Sweep as Sweep

from .uncalculated_sweep import UncalculatedSweep as UncalculatedSweep
from .uncalculated_sweep import ApriorUncalculatedSweep as ApriorUncalculatedSweep
from .spectrogram import Spectrogram as Spectrogram

from .config.base_config import Config as Config
from .config.sweep_config import SweepConfig as SweepConfig



from typing import Callable

import numpy as np


def f_t_linear_function(time_start=0., time_end=10., f_start=1., f_end=100.) \
        -> Callable[[np.ndarray], np.ndarray]:
    '''Create liner function of changes frequency-time.

    Use parameters to calculate linear function f(t) = b*t+k, where t is time,
    f(t) function frequency-time.

    Args:
        time_start (RealNumber, optional): start time. Defaults to 0..
        time_end (RealNumber, optional): end time. Defaults to 10..
        f_start (RealNumber, optional): start frequency. Defaults to 1..
        f_end (RealNumber, optional): end frequency. Defaults to 100..

    Returns:
        Callable[[np.ndarray], np.ndarray]: linear function frequency-time.
    '''
    return lambda t: t * (f_end - f_start) / (time_end - time_start) + f_start


def f_t_linear_array(time: np.ndarray, f_start=1., f_end=100.) -> np.ndarray:
    '''Create array of number describe linear changes frequency-time.

    Use parameters to calculate linear function f(t) = b*t+k, where t is time,
    f(t) function frequency-time.

    Args:
        time (np.ndarray): time changes array.
        f_start (RealNumber, optional): start frequency. Defaults to 1..
        f_end (RealNumber, optional): end frequency. Defaults to 100..

    Returns:
        np.ndarray: array of numbers describe linear changes frequency-time.
    '''
    return time * (f_end - f_start) / (time[-1] - time[0]) + f_start



from typing import Any, List

from PyEMD import CEEMDAN, EMD  # type: ignore
from ..signal import Signal


def get_IMFs_ceemdan(
    data: Signal,
    number_seedman=30,
    epsilon=0.005,
    ext_EMD: Any = None,
    parallel=False,
    processes=None,
    noise_scale=1,
    noise_kind="normal",
    range_thr=0.01,
    total_power_thr=0.05,
) -> List[Signal]:
    '''Empirical mode decomposition (EMD).
    Using CEEMDAN from PyEMD (https://pyemd.readthedocs.io/) to calculate IMFs
    and return them as a list of `Signals`.

    About empirical mode decomposition on
    https://en.wikipedia.org/wiki/Hilbert%E2%80%93Huang_transform#Techniques

    Args:
        data (Signal): _description_

        number_seedman (int, optional): _description_. Defaults to 30.

        epsilon (float, optional): Scale for added noise (\\epsilon) which
            multiply std \\sigma: \beta = \\epsilon \\cdot \\sigma.
            Defaults to 0.005.

        ext_EMD (Any, optional): One can pass EMD object defined outside,
            which will be used to compute IMF decompositions in each trial.
            If none is passed then EMD with default options is used.
            Defaults to None.

        parallel (bool, optional): Flag whether to use multiprocessing in
            EEMD execution. Since each EMD(s+noise) is independent this should
            improve execution speed considerably. *Note* that it's disabled by
            default because it's the most common problem when CEEMDAN takes too
            long time to finish. If you set the flag to True, make also sure to
            set processes to some reasonable value. Defaults to False.

        processes (_type_, optional): Number of processes harness when
            executing in parallel mode. The value should be between 1 and
            max that depends on your hardware. Defaults to None.

        noise_scale (int, optional): Scale (amplitude) of the added noise.
            Defaults to 1.

        noise_kind (str, optional):  What type of noise to add. Allowed
            are "normal" (default) and "uniform". Defaults to "normal".

        range_thr (float, optional):  Range threshold used as an IMF check.
            The value is in percentage compared to initial signal's amplitude.
            If absolute amplitude (max - min) is below the range_thr then the
            decomposition is finished. Defaults to 0.01.

        total_power_thr (float, optional): Signal's power threshold.
            Finishes decomposition if sum(abs(r)) < thr. Defaults to 0.05.

    Returns:
        List[Signal]: List of Signals expected IMFs.
    '''
    y = data.y
    emd = CEEMDAN(
        number_seedman,
        epsilon,
        ext_EMD,
        parallel,
        processes=processes,
        noise_scale=noise_scale,
        noise_kind=noise_kind,
        range_thr=range_thr,
        total_power_thr=total_power_thr,
    )
    IMFs = emd(y)
    result = [Signal(data.x.copy(), k) for k in IMFs]
    return result


def get_IMFs_emd(data: Signal) -> List[Signal]:
    '''"Empirical mode decomposition (EMD).

    Using EMD from PyEMD (https://pyemd.readthedocs.io/) to calculate IMFs
    and return them as a list of `Signals`.

    About empirical mode decomposition on
    https://en.wikipedia.org/wiki/Hilbert%E2%80%93Huang_transform#Techniques

    Args:
        data (Signal): signal to calculate IMFs.

    Returns:
        List[Signal]: List of Signals expected IMFs.
    '''
    y = data.y
    emd = EMD()
    IMFs = emd(y)
    result = [Signal(data.x.copy(), k) for k in IMFs]
    return result



from typing import Union
import numpy as np

from ..axis import ArrayAxis, get_array_axis_from_array

from ..relation import Relation
from ..sweep import Sweep
from ..uncalculated_sweep import ApriorUncalculatedSweep
from ..utility_functions.ftat_functions import dwell
from ..utility_functions.a_t import tukey_a_t


def get_dwell_sweep(
    time: Union[ArrayAxis, np.ndarray],
    f_start=1.0,
    f_central=5.0,
    f_end=100.0,
    time_tapper=1.0,
    aprior_data: Relation = None,
) -> Sweep:
    """Create a dwell sweep using an a priori data.

    time_tapper in seconds is used to apply tukey function at the end of dwell sweep signal.
    """

    if not isinstance(time, ArrayAxis):
        time = get_array_axis_from_array(time)

    ftat_method = dwell(f_start, f_end, f_central)
    uasw = ApriorUncalculatedSweep(time, aprior_data, ftat_method)
    tukey_array = tukey_a_t(time.array, time_tapper, "both")
    tukey_window = Relation(time, tukey_array)

    return uasw() * tukey_window



import typing
from typing import Any, Callable, Tuple, Union, TYPE_CHECKING

import numpy as np
from packaging import version

RealNumber = Union[float, int]
'''A real number.'''

Number = Union[float, int, complex]
'''A real or complex number'''

Frequency = np.ndarray
'''The frequency expected by numpy array.'''

Time = np.ndarray
'''The time expected by numpy array.'''

ImageSpectrogram = np.ndarray
'''The image spectrogram expected by numpy 2D array.'''

Envelope = np.ndarray
'''The envelop expected by numpy array.'''

Spectrogram = Tuple[Time, Frequency, ImageSpectrogram]
'''The spectrogram gathering time, frequency and image spectrogram in one
object Tuple.'''


Theta = "relation.Relation"
'''`Relation` of phase.'''

Ftat = Union[np.ndarray, Callable[[np.ndarray], np.ndarray]]
'''The Representation of object from which will be extracted frequency
modulation or amplitude modulation.'''

X = np.ndarray
'''x array'''
Y = np.ndarray
'''y array'''

if TYPE_CHECKING:
    from typing import Literal
else:
    if hasattr(typing, 'Literal'):
        from typing import Literal
    else:
        from typing_extensions import Literal


def get_type():
    '''Get type of array like object from numpy depends on version of numpy.

    Returns:
        _type_: type of array like object.
    '''
    if version.parse(np.__version__) > version.parse("1.19"):
        from numpy.typing import ArrayLike as NpArrayLike
        return NpArrayLike
    else:
        return Any


ArrayLike = get_type()



from typing import Union

import numpy as np

from .axis import ArrayAxis
from .config.sweep_config import SweepConfig
from .config.base_config import Config
from .core import RelationProtocol
from .defaults.sweep_methods import Spectrogram as DataSpectrogram
from .relation import Relation
from .signal import Signal
from .spectrogram import Spectrogram
from .help_types import ArrayLike


class Sweep(Signal):
    '''Class `Sweep`.

    A class for analyzing changes in a signal over time, inherited from
    the `Signal` class.

    For analysis, you can use not only the sweep signal, but also other
    signals for which the spectrogram needs to be considered.

    When creating an instance of the class, the spectrogram is calculated.
    The method used to calculate the spectrogram is defined in the
    `SweepConfig` class. You can override it with your own.

    If the frequency vs. time and amplitude vs. time functions have not
    been passed, they are also calculated, the `get_frequency_time`,
    `get_amplitude_time` methods defined in the `SweepConfig` class are used.

    Perform the same operations as for the inherited class.

    '''

    def __init__(
        self,
        time: Union[RelationProtocol, ArrayAxis, ArrayLike],
        amplitude: ArrayLike = None,
        frequency_time: Relation = None,
        amplitude_time: Relation = None,
        a_prior_signal: Signal = None,
    ) -> None:
        '''Initialize sweep instance.

        Args:
            time (Union[RelationProtocol, ArrayAxis, ArrayLike]):
                The `Relation` class, or a class derived from the Relation class,
                or ArrayAxis, or an array_like object containing
                numbers(real or complex).

            amplitude (ArrayLike, optional): `None` or array_like object
                containing real or complex numbers. Defaults to None.

            frequency_time (Relation, optional):
                This parameter describes the change in frequency versus time of the
                transmitted signal. Defaults to None.

            amplitude_time (Relation, optional): This parameter describes
                the change in amplitude envelop of signal from the time of the
                transmitted signal. Defaults to None.

            a_prior_signal (Signal, optional): The signal used to create
                the sweep signal. Defaults to None.

        '''

        super().__init__(time, amplitude)

        self.frequency_time = (
            frequency_time
            if frequency_time is not None
            else SweepConfig.get_f_t(self)
        )
        self.amplitude_time = (
            amplitude_time
            if amplitude_time is not None
            else SweepConfig.get_a_t(self)
        )

        spectrogram = SweepConfig.spectrogram_method(self)

        self.spectrogram = _get_spectrogram(spectrogram)

        self.a_prior_signal = a_prior_signal


def _get_spectrogram(spectrogram: DataSpectrogram) -> Spectrogram:
    spectrogram_ = spectrogram[2]
    if spectrogram[0].size < 2:
        time = np.append(spectrogram[0]
                         [0], spectrogram[0][0] + 1)
        spectrogram_ = np.concatenate(
            (
                spectrogram_,
                np.zeros((spectrogram_.shape[0], 1))
            ),
            axis=1)
    else:
        time = spectrogram[0]

    if spectrogram[1].size < 2:
        frequency = np.append(
            spectrogram[1][0], spectrogram[1][0] + 1)
        spectrogram_ = np.concatenate(
            (
                spectrogram_,
                np.zeros((1, spectrogram_.shape[1]))
            ),
            axis=0)
    else:
        frequency = spectrogram[1]

    return Spectrogram(
        time=Config.get_array_axis_from_array_method(time),
        frequency=Config.get_array_axis_from_array_method(frequency),
        spectrogram=spectrogram[2],
    )



from typing import Union
import math
import random

import numpy as np
from scipy.signal import max_len_seq  # type: ignore

from ..relation import Relation
from ..axis import ArrayAxis, get_array_axis_from_array


def get_m_sequence_code(length_code: int, is_full=False) -> np.ndarray:
    '''Create m-sequence array

    Args:
        length_code (int): length of m-sequence array.

        is_full (bool, optional): parameter to return full sequence if True.
            sequence can be longer then requested length code. If False return
            sequence which size equal length_code.
            Defaults to False.

    Returns:
        np.ndarray: array of m-sequence.
    '''
    len_sequence = math.ceil(math.log((length_code - 1), 2))
    start_sequence = np.array([random.randint(0, 1)
                              for _ in range(len_sequence)])

    if np.all(start_sequence == 0):
        start_sequence[0] = 1

    m_sequence = max_len_seq(nbits=len_sequence, state=start_sequence)[0]

    if is_full:
        return m_sequence

    return m_sequence[:length_code - 1]


def get_relation_m_sequence(time: Union[ArrayAxis, np.ndarray],
                            start_sequence: np.ndarray = None,
                            is_full=False) -> Relation:
    '''Create m-sequence relation.

    Args:
        time (Union[ArrayAxis, np.ndarray]): time of sequence.

        start_sequence (np.ndarray, optional): start array sequence of -1 and 1
            to create m-sequence. Defaults to None.

        is_full (bool, optional): if True return full m-sequence with new
            time with equal size to m-sequence array. Defaults to False.

    Returns:
        Relation: relation of m-sequence.
    '''
    if not isinstance(time, ArrayAxis):
        time = get_array_axis_from_array(time)

    len_seq = math.ceil(math.log((time.size - 1), 2))

    if start_sequence is None:
        start_sequence = np.array([random.randint(0, 1)
                                  for _ in range(len_seq)])

    if np.all(start_sequence == 0):
        start_sequence[0] = 1

    m_sequence = max_len_seq(
        nbits=start_sequence.size,
        state=start_sequence)[0]

    if is_full:
        new_time = ArrayAxis(
            0, time.sample * (m_sequence.size - 1), time.sample)
        return Relation(new_time, m_sequence)

    return Relation(time, m_sequence[:time.size])



from ..defaults import sweep_methods as dfsm


class SweepConfig:

    """The configuration file for sweep.

    Several methods are defined for default calculations:

    `spectrogram_method_default`:

    The method by which the spectrogram will be calculated.
    Method derived from default function:
    `sweep_design.defaults.sweep_methods.get_spectrogram`

    Args:
        sweep (relation.Relation): instance of sweep signal.

    Returns:
        Spectrogram: tuple of np.ndarray. The first element is time.
            The second is frequency. The third is matrix M x N of spectrogram.

    ---

    `get_f_t`:

    The method by which frequency versus time will be calculated.
    Method derived from default function:
    `sweep_design.defaults.sweep_methods.get_f_t`

    Args:
        sweep (Relation): instance of sweep signal.

    Returns:
        Relation: instance `Relation`

    ---

    `get_a_t`:

    The method by which the time envelope of the signal will be calculated.
    Method derived from default function:
    `sweep_design.defaults.sweep_methods.get_a_t`

    Args:
        sweep (Relation): instance of sweep signal.

    Returns:
        Relation: instance `Relation`

    ---

    `freq2time`:

    The simple method to extract the time envelope of a sweep signal and
    the time-frequency function to generate a sweep signal from a priori data.
    Method derived from default function:
    `sweep_design.defaults.sweep_methods.simple_freq2time`

    Args:
        spectrum (Spectrum): instance signal of 'Spectrum'

    Returns:
        Tuple[Time, Frequency, Envelope]: simple representation Frequency
            modulation from a prior spectrum.

    ---

    The above methods can be overridden with your own here, or you can import the
    class SweepConfig somewhere and override it there.
    (They must be written according to the rules corresponding to
    the input and output parameters)
    """

    # Methods for Sweep.
    spectrogram_method = dfsm.get_spectrogram
    get_f_t = dfsm.get_f_t
    get_a_t = dfsm.get_a_t

    # Method for ApriorUncalculatedSweep.
    freq2time = dfsm.simple_freq2time



import numpy as np

from typing import Union
from ..axis import ArrayAxis
from ..sweep import Sweep
from ..uncalculated_sweep import UncalculatedSweep
from ..utility_functions.f_t import f_t_linear_array
from ..utility_functions.a_t import tukey_a_t


def get_linear_sweep(time: Union[ArrayAxis, np.ndarray], f_start=1.0,
                     f_end=100.0, time_tapper=None) -> Sweep:
    """Create linear sweep.

    time_tapper in seconds is used to apply tukey function to a sweep signal.
    """

    time_array = time if isinstance(time, np.ndarray) else time.array

    f_t = f_t_linear_array(time_array, f_start, f_end)
    a_t = tukey_a_t(time_array, time_tapper)

    unsw = UncalculatedSweep(time, f_t, a_t)
    return unsw()



from abc import ABC, abstractmethod
from enum import Enum
from typing import Tuple

import numpy as np

from .axis import ArrayAxis


class RelationProtocol(ABC):
    '''Abstract class how create Relation classes.

    '''

    @property
    @abstractmethod
    def array(self) -> np.ndarray:
        pass

    @property
    @abstractmethod
    def x(self) -> ArrayAxis:
        pass

    @property
    @abstractmethod
    def y(self) -> np.ndarray:
        pass

    @abstractmethod
    def get_data(self) -> Tuple[np.ndarray, np.ndarray]:
        pass


class MathOperation(Enum):
    '''Which is math operations will be used.

    Inheritance `Enum` class.

    Args:
        Enum (_type_): base `Enum` class
    '''
    ADD = "__add__"
    RADD = "__radd__"
    SUB = "__sub__"
    RSUB = "__rsub__"
    MUL = "__mul__"
    RMUL = "__rmul__"
    TRUEDIV = "__truediv__"
    RTRUEDIV = "__rtruediv__"
    POW = "__pow__"
    RPOW = "__rpow__"



from typing import Optional, Type, TypeVar, Union

import numpy as np

from . import signal
from .axis import ArrayAxis
from .config.base_config import Config
from .core import RelationProtocol
from .exc import ConvertingError
from .relation import Relation
from .help_types import ArrayLike, Number

SP = TypeVar("SP", bound="Spectrum")
'''Instance of `Spectrum`'''

SSPR = Union["Spectrum", "signal.Signal", Relation]
'''Instance of Spectrum or Signal or Relation'''

SSPRN = Union["Spectrum", "signal.Signal", Relation, Number]
'''Instance of Spectrum or Signal or Relation or Number'''


class Spectrum(Relation):
    '''A class that describes the spectrum of a signal.

    The `Spectrum` class derived from the `Relation` class.

    Each 'Spectrum' can be converted into a `Signal` using method `get_spectrum`
    To convert the `Spectrum` into a `Signal`, the method defined in the `Config`
    class is used. (`Config.spectrum2signal_method`). Current method can be
    overridden by own in `Config` class.

    When performing arithmetic operations on instances of the `Signal` class,
    an instance of the `Spectrum` class will be extracted from
    the `Signal` instance, and arithmetic operations will be performed
    on this instance. An instance of `Relation` class will be converted into
    the instance of `Spectrum` class.

    '''

    def __init__(
        self,
        frequency: Union[RelationProtocol, ArrayAxis, ArrayLike],
        spectrum_amplitude: ArrayLike = None,

    ) -> None:
        '''Initialization of instance of `Spectrum`.

        Args:
            frequency (Union[RelationProtocol, ArrayAxis, ArrayLike]):
                An instance of `Relation` class or inherited from it,
                or `ArrayAxis` instance, or array_like object containing
                numbers (real or complex).

            spectrum_amplitude (ArrayLike, optional):
                None or array_like object containing numbers (real or complex).
                Defaults to None.

        '''
        super().__init__(frequency, spectrum_amplitude)
        self._spectrum2signal_method_default = Config.spectrum2signal_method
        self.signal: Optional[signal.Signal] = None

    @property
    def frequency(self) -> ArrayAxis:
        '''Frequency array axis.

        Equal to property `x`.
        Returns:
            ArrayAxis: frequency array axis.
        '''
        return self.x

    @property
    def amplitude(self) -> np.ndarray:
        '''Spectrum amplitude array.

        Equal to property `y`.
        Returns:
            np.ndarray: spectrum amplitude array.
        '''
        return self.y

    def get_signal(self, recalculate=False,
                   start_time: float = None) -> "signal.Signal":
        '''Get signal from spectrum.

        Compute the signal from the spectrum.

        Args:
            recalculate (bool, optional): If `True` then the spectrum will be
                calculated again, else false, then the saved one will be used.
                Defaults to `False`. Defaults to False.

            start_time (float, optional): If True then the signal will be
                shifted to zero. Defaults to `False`.

        Returns:
            signal.Signal: instance of `Signal` described this `Spectrum`.
        '''

        if self.signal is None or recalculate:
            time, amplitude = self._spectrum2signal_method_default(
                self, start_time
            )
            self.signal = signal.Signal(time, amplitude)
        return self.signal

    def get_amp_spectrum(self: SP) -> Relation:
        '''Get amplitude spectrum.

        Calculate the relationship between the frequency and the absolute
        value of the spectrum amplitude.

        Args:
            self (SP): instance of Spectrum.

        Returns:
            Relation: new instance of Relation.
        '''
        return Relation(self.x.copy(), np.abs(self.y))

    def get_phase_spectrum(self: SP) -> Relation:
        '''Calculate the relationship between frequency and phase of the spectrum.

        Args:
            self (SP): instance of Spectrum.

        Returns:
            Relation: new instance of Relation.
        '''

        return Relation(self.x.copy(), np.unwrap(np.angle(self.y)))

    def get_reverse_filter(
        self: SP,
        percent: Union[float, int] = 5.0,
        subtrack_phase=True,
        frequency_start: float = None,
        frequency_end: float = None,
    ) -> SP:
        '''Calculate filter of reversed signal.

        Args:
            self (SP): instance of Spectrum.

            percent (Union[float, int], optional): level of added white noise
                in percent. Defaults to 5.0.

            subtrack_phase (bool, optional): If True performs phase subtraction.
                If False succeeds, add the phase. Defaults to True.

            frequency_start (float, optional): The start frequency. Defaults to None.
            frequency_end (float, optional): The end frequency. Defaults to None.

        Returns:
            SP: new instance of Spectrum.
        '''

        spectrum = self.select_data(frequency_start, frequency_end)
        abs_spectrum = spectrum.get_amp_spectrum()
        abs_spectrum = abs_spectrum + abs_spectrum.max() * percent / 100
        reversed_abs_spectrum = 1 / abs_spectrum

        if subtrack_phase:
            phase_spectrum = -1 * spectrum.get_phase_spectrum()
        else:
            phase_spectrum = 1 * spectrum.get_phase_spectrum()

        result_spectrum = type(self).get_spectrum_from_amp_phase(
            reversed_abs_spectrum, phase_spectrum
        )
        return result_spectrum

    def add_phase(self: SP, other: SSPR) -> SP:
        '''Add phase to spectrum.

        Args:
            self (SP): instance of `Spectrum`

            other (SSPR): Extracting the `Spectrum` from the object and adding
                the phase `Spectrum` to the `Spectrum`.

        Returns:
            `SP`: new instance of `Spectrum`.
        '''

        sp_other = _input2spectrum(other)
        return type(self).get_spectrum_from_amp_phase(
            self.get_amp_spectrum(),
            self.get_phase_spectrum() + sp_other.get_phase_spectrum(),

        )

    def sub_phase(self: SP, other: SSPR) -> SP:
        '''Subtrack phase from spectrum.

        Args:
            self (SP): instance of `Spectrum`

            other (SSPR): Extracting the `Spectrum` from the object and subtrack
                the phase `Spectrum` from the `Spectrum`.

        Returns:
            `SP`: new instance of `Spectrum`.
        '''
        sp_other = _input2spectrum(other)
        return type(self).get_spectrum_from_amp_phase(
            self.get_amp_spectrum(),
            self.get_phase_spectrum() - sp_other.get_phase_spectrum(),

        )

    @classmethod
    def get_spectrum_from_amp_phase(
        cls: Type[SP], amplitude_spectrum: Relation, phase_spectrum: Relation
    ) -> SP:
        '''Calculate of the spectrum from the amplitude and phase spectrum.

        The spectrum is calculated through the amplitude and phase spectrum
        using the formula abs*exp(1j*phase).

        Args:
            cls (Type[SP]): Spectrum class.
            amplitude_spectrum (Relation): Amplitude spectrum is instance of 'Relation'.
            phase_spectrum (Relation): Phase spectrum is instance of 'Relation'.

        Returns:
            SP: new instance of `Spectrum`
        '''

        return cls(amplitude_spectrum *
                   ((1.0j * phase_spectrum).exp()))

    @classmethod
    def convolve(cls: Type[SP], r1: SSPR, r2: SSPR) -> SP:
        '''Convolution of two instances of `Relation` and return new instance of
        `Spectrum`. Instances of `Signal` will be converted to `Spectrum`

        Args:
            cls (Type[S]): `Signal` class.
            r1 (SSPR): instance of `Relation` or subclass of 'Relation'
            r2 (SSPR): instance of `Relation` or subclass of 'Relation'

        Returns:
            S: new instance of `Spectrum`.
        '''
        sp_r1 = _input2spectrum(r1)
        sp_r2 = _input2spectrum(r2)
        return super().convolve(sp_r1, sp_r2)

    @classmethod
    def correlate(cls: Type[SP], r1: SSPR, r2: SSPR) -> SP:
        '''Correlation of two instances of `Relation` and return new instance of
        `Spectrum`. Instances of `Signal` will be converted to `Spectrum`

        Args:
            cls (Type[S]): `Signal` class.
            r1 (SSPR): instance of `Relation` or subclass of 'Relation'
            r2 (SSPR): instance of `Relation` or subclass of 'Relation'

        Returns:
            S: new instance of `Spectrum`.
        '''
        sp_r1 = _input2spectrum(r1)
        sp_r2 = _input2spectrum(r2)
        return super().correlate(sp_r1, sp_r2)

    def __add__(self: SP, a: SSPRN) -> SP:
        r_a = _input2spectrum_operation(a)
        return super().__add__(r_a)

    def __sub__(self: SP, a: SSPRN) -> SP:
        r_a = _input2spectrum_operation(a)
        return super().__sub__(r_a)

    def __mul__(self: SP, a: SSPRN) -> SP:
        r_a = _input2spectrum_operation(a)
        return super().__mul__(r_a)

    def __truediv__(self: SP, a: SSPRN) -> SP:
        r_a = _input2spectrum_operation(a)
        return super().__truediv__(r_a)

    def __pow__(self: SP, a: SSPRN) -> SP:
        r_a = _input2spectrum_operation(a)
        return super().__pow__(r_a)


def _input2spectrum_operation(
        inp: SSPRN) -> Union[Relation, Spectrum, Number]:
    if isinstance(inp, signal.Signal):
        return inp.get_spectrum()
    elif isinstance(inp, (Spectrum, Relation, int, float, complex)):
        return inp
    else:
        raise ConvertingError(type(inp), Spectrum)


def _input2spectrum(inp: SSPR) -> Spectrum:
    if isinstance(inp, signal.Signal):
        return inp.get_spectrum()

    elif isinstance(inp, Spectrum):
        return inp

    elif isinstance(inp, Relation):
        return Spectrum(inp)
    else:
        raise ConvertingError(type(inp), Spectrum)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sweep_design',
 'sweep_design.config',
 'sweep_design.defaults',
 'sweep_design.prepared_sweeps',
 'sweep_design.utility_functions']

package_data = \
{'': ['*']}

install_requires = \
['emd-signal>=1.2.3,<2.0.0', 'packaging>=22,<23']

extras_require = \
{':python_version >= "3.6" and python_version < "3.8"': ['numpy<=1.19',
                                                         'scipy>=1.2,<1.8',
                                                         'typing_extensions>=3.6'],
 ':python_version >= "3.8"': ['numpy>1.19', 'scipy>=1.8']}

setup_kwargs = {
    'name': 'sweep-design',
    'version': '0.2.0',
    'description': 'Project designed to create and analyze sweep signals.',
    'long_description': '# sweep-design\n\nSimple way to create sweep signal.\n\n## The project is intended for designing sweep signals.\n\nThe package is intended to create and develop sweep signals of\nvarying complexity.\n\nThe project can be used both for educational and work purposes.\n\nIt is convenient to use [`Jupyter Lab`](https://jupyter.org/) or\n[`Jupyter Notebook`](https://jupyter.org/) to speed up the development\nof signals, to compare their parameters with other signals,\nand to visualize them.\n\nThe project is designed so that you can easily change the creation of sweep\nsignals. For example, write your own methods describing how the frequency\nand amplitude will change from the time of the sweep signal.\n\nThe project was made to be able to create various sweep signals: implemented\nand not implemented by a vibration source, from simple ones, like a linear\nsweep signal, to complex ones, like a pseudo-random sweep signal.\n\nTools have been written with which unrealizable sweep signals\ncould be made realizable.\n\nIn addition, documentation consist tutorial how to work with library\nand examples of ready-made sweeps. You can write own sweep creation.\n\n# Installation\n\nTo install use:\n\n```bash\n$ pip install sweep-design\n```\n\nor using `poetry`\n\n```bash\n$ poetry add sweep-design\n```\n\nAlso you can clone or load project from [GitHub](https://github.com/Omnivanitate/sweep-design),\nand install requirement packages using the\n\n```bash\n$ pip install -r requirement.txt\n```\n\nor if you want develop, use\n\n```bash\n$ pip install -r requirement-dev.txt\n```\n\nor\n\n```bash\n$ poetry install\n```\n\nor coping pieces of code and create your own.\n\n## Usage\n\nThe project is a library. Working with it is the same as with\nother third-part libraries of the python language.  \nAn example of how to include the library is described\n[here](https://docs.python.org/3/tutorial/modules.html).\n\nThe library consists sub-modules:\n\n- `sweep_design.config` - contains the project configuration `Config` and `SweepConfig`.\n- `sweep_design.defaults` - contains default methods to calculate.\n- `sweep_design.prepared_sweep` - contains sweep signal templates.\n- `sweep_design.utility_functions` - contains function to work with signals.\n- `sweep_design.core` - contains basic classes `MathOperation` and `RelationProtocol`.\n- `sweep_design.exc` - contains exceptions.\n- `sweep_design.axis` - contains class `ArrayAxis`\n- `sweep_design.relation ` - contains class `Relation`\n- `sweep_design.signal ` - contains class `Signal`\n- `sweep_design.spectrum ` - contains class `Spectrum`\n- `sweep_design.sweep` - contains class `Sweep`\n- `sweep_design.uncalculated` - contains classes `UncalculatedSweep` and `ApriorUncalculatedSweep`\n- `sweep_design.spectrogram` - contains classes `Spectrogram`\n\nFor convenient base classes:\n`ArrayAxis`, `Relation`, `Signal`, `Spectrum`, `Sweep`, `UncalculatedSweep`,\n`ApriorUncalculatedSweep`, `Config`, `ConfigSweep` - can be imported from\na `sweep_design` module.\n\nFor example:\n\n```python\nfrom sweep_design import Signal\n```\n\nUtility functions can be imported from `sweep_design.utility_functions`.  \nAnd prepared sweep - from `sweep_design.prepared_sweep`.\n\n### Quick start. Simple work flow.\n\nBelow is a simple example of creating a sweep signal and visualizing it.\nA more extended description of the work of the library in the documentation.\nOther examples are contained in the examples contains in _Tutorial_ and\n_Prepared sweep_ sections.\n\nFor the following code [`Matplotlib`](https://matplotlib.org/) need be used\nto visualize a result of work. But `Matplotlib` can be replaced with another\nlibrary that you use.\n\n```python\nimport matplotlib.pyplot as plt\n\nfrom sweep_design import ArrayAxis, UncalculatedSweep\n\ntime = ArrayAxis(start=0., end=10., sample=0.01)\n\nusw = UncalculatedSweep(time=time)\nsw = usw()\n\nt_sw, a_sw = sw.get_data()\nplt.plot(t_sw, a_sw)\nplt.xlabel(\'Time, s\')\nplt.ylabel(\'Amplitude\')\nplt.title(\'Sweep-signal\')\n```\n\nResult:\n\n![sweep_with_matplotlib](https://user-images.githubusercontent.com/89973180/156033978-ccc8de40-9f6b-4bb1-b59f-7a3ea41d2f64.png "Linear Sweep")\n\n## Credits\n\n`sweep-design` was created with  \n[`numpy`](https://numpy.org/)  \n[`scipy`](https://scipy.org/)  \n[`EMD-signal`](https://pyemd.readthedocs.io/en/latest/)\n',
    'author': 'Vladislav',
    'author_email': 'serebraykov.vs@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import Optional, Type, TypeVar, Union

import numpy as np  # type: ignore

from . import spectrum
from .axis import ArrayAxis
from .config.base_config import Config
from .core import RelationProtocol
from .exc import ConvertingError
from .relation import Relation
from .help_types import ArrayLike, Number

S = TypeVar("S", bound="Signal")
'''Instance of `Signal`.'''

SSPR = Union["spectrum.Spectrum", "Signal", Relation]
'''Instance of `Signal` or `Spectrum` or `Relation`.'''

SSPRN = Union["spectrum.Spectrum", "Signal", Relation, Number]
'''Instance of `Signal` or `Spectrum` or `Relation` or `Number`.'''


class Signal(Relation):
    '''Class describing some kind of signal.

    The `Signal` class inherits the `Relation` class.

    Each signal can be converted into a `spectrum.Spectrum` using method `get_spectrum`.
    To convert the signal into a spectrum, the method defined in the `Config`
    class is used. (Config.signal2spectrum_method). Current method can be
    overridden by own in `Config` class.

    When performing arithmetic operations on instances of the spectrum.Spectrum class,
    an instance of the `Signal` class will be extracted from
    the `spectrum.Spectrum` instance, and arithmetic operations will be performed
    on this instance. An instance of `Relation` class will be converted into
    the instance of `Signal` class.
    '''

    def __init__(
        self,
        time: Union[RelationProtocol, ArrayAxis, ArrayLike],
        amplitude: ArrayLike = None,
        **kwargs
    ) -> None:
        '''Initialization of instance of `Signal`.

        Args:
            time (Union[RelationProtocol, ArrayAxis, ArrayLike]): An instance
                of Relation class or inherited from it, or ArrayLike, or array_like
                object containing numbers (real or complex).

            amplitude (ArrayLike, optional): None or array_like object
                containing numbers (real or complex). Defaults to None.
        '''

        self._signal2spectrum_method_default = Config.signal2spectrum_method
        super().__init__(time, amplitude, **kwargs)
        self._spectrum: Optional[spectrum.Spectrum] = None

    @property
    def time(self) -> ArrayAxis:
        '''Time array axis.

        Equal to property `x`.

        Returns:
            ArrayAxis: time array axis.
        '''
        return self.x

    @property
    def amplitude(self) -> np.ndarray:
        '''Amplitude array.

        Equal to property 'y'

        Returns:
            np.ndarray: amplitude array.
        '''
        return self.y

    def get_spectrum(self, recalculate=False,
                     is_start_zero=False) -> "spectrum.Spectrum":
        '''Get spectrum from signal.

        Args:
            recalculate (bool, optional): If `True` then the spectrum will be
                calculated again, else false, then the saved one will be used.
                Defaults to `False`.

            is_start_zero (bool, optional): If True then the signal will be
                shifted to zero. Defaults to `False`.

        Returns:
            spectrum.Spectrum: instance of `spectrum.Spectrum` described this `Signal`.
        '''

        if self._spectrum is None or recalculate:
            f, a = self._signal2spectrum_method_default(
                self, is_start_zero)
            self._spectrum = spectrum.Spectrum(f, a)

        return self._spectrum

    def get_amplitude_spectrum(
        self, recalculate=False, is_start_zero=False
    ) -> Relation:
        '''Extract amplitude spectrum from `spectrum.Spectrum`.

        Method `get_spectrum` is used to get instance of `spectrum.Spectrum`.
        The amplitude spectrum is calculated from it using `get_amp_spectrum`
        method.

        Args:
            recalculate (bool, optional): If `True` then the spectrum will be
                calculated again, else false, then the saved one will be used.
                Defaults to `False`.

            is_start_zero (bool, optional): If True then the signal will be
                shifted to zero. Defaults to `False`.

        Returns:
            Relation: amplitude spectrum expected Relation instance.
        '''
        return self.get_spectrum(recalculate, is_start_zero).get_amp_spectrum()

    def get_phase_spectrum(self, recalculate=False,
                           is_start_zero=False) -> Relation:
        '''Extract amplitude spectrum from `spectrum.Spectrum`.

        Method `get_spectrum` is used to get instance of `spectrum.Spectrum`.
        The amplitude spectrum is calculated from it using `get_amp_spectrum`
        method.

        Args:
            recalculate (bool, optional): If `True` then the spectrum will be
                calculated again, else false, then the saved one will be used.
                Defaults to `False`.

            is_start_zero (bool, optional): If True then the signal will be
                shifted to zero. Defaults to `False`.

        Returns:
            Relation: amplitude spectrum expected Relation instance.
        '''
        return self.get_spectrum(
            recalculate, is_start_zero).get_phase_spectrum()

    def get_reverse_signal(
        self: S,
        percent: Union[float, int] = 5.0,
        subtract_phase: bool = True,
        frequency_start: float = None,
        frequency_end: float = None,
    ) -> S:
        '''Calculate reversed signal.

        Args:
            self (S): instance of Signal.

            percent (Union[float, int], optional): level of added white
                noise in percent. Defaults to 5.0.

            subtract_phase (bool, optional): If True performs phase subtraction,
                If False succeeds, add the phase. Defaults to True.

            frequency_start (float, optional): The start frequency.
                Defaults to None.

            frequency_end (float, optional): The end frequency.
                Defaults to None.

        Returns:
            S: instance of Signal.
        '''

        signal = (
            self.get_spectrum()
            .get_reverse_filter(
                percent,
                subtract_phase,
                frequency_start,
                frequency_end
            )
            .get_signal()
        )

        return type(self)(signal)

    def add_phase(self: S, other: SSPR) -> S:
        '''Add phase to signal.

        Args:
            self (S): instance of `Signal`.

            other (SSPR): Extracting the `spectrum.Spectrum` from the object and adding
                the phase `spectrum.Spectrum` to the `Signal`.

        Returns:
            S: new instance of `Signal`.
        '''
        sp_other = spectrum._input2spectrum(other)
        self_spectrum = self.get_spectrum()
        new_spectrum = spectrum.Spectrum.get_spectrum_from_amp_phase(
            self_spectrum.get_amp_spectrum(),
            self_spectrum.get_phase_spectrum() + sp_other.get_phase_spectrum(),
        )
        return type(self)(new_spectrum.get_signal())

    def sub_phase(self: S, other: SSPR) -> S:
        '''Subtrack phase from signal.

        Args:
            self (S): instance of Signal

            other (SSPR): Extracting the spectrum from the object and subtrack
                the phase spectrum from the signal.

        Returns:
            S: new instance of Signal.
        '''
        sp_other = spectrum._input2spectrum(other)
        self_spectrum = self.get_spectrum()
        new_spectrum = spectrum.Spectrum.get_spectrum_from_amp_phase(
            self_spectrum.get_amp_spectrum(),
            self_spectrum.get_phase_spectrum() - sp_other.get_phase_spectrum(),
        )
        return type(self)(new_spectrum.get_signal())

    @classmethod
    def convolve(cls: Type[S], r1: SSPR, r2: SSPR) -> S:
        '''Convolution of two instances of `Relation` and return new instance of
        `Signal`. Instances of `spectrum.Spectrum` will be converted to `Signal`

        Args:
            cls (Type[S]): `Signal` class.
            r1 (SSPR): instance of `Relation` or subclass of 'Relation'
            r2 (SSPR): instance of `Relation` or subclass of 'Relation'

        Returns:
            S: new instance of Signal.
        '''
        s_r1 = _inp2signal(r1)
        s_r2 = _inp2signal(r2)
        return cls(super().convolve(s_r1, s_r2))

    @classmethod
    def correlate(cls: Type[S], r1: SSPR, r2: SSPR) -> S:
        '''Correlation of two instances of `Relation` and return new instance of
        `Signal`. Instance of `spectrum.Spectrum` will be converted to `Signal`

        Args:
            cls (Type[S]): `Signal` class.
            r1 (SSPR): instance of `Relation` or subclass of 'Relation'
            r2 (SSPR): instance of `Relation` or subclass of 'Relation'

        Returns:
            S: new instance of Signal.
        '''
        s_r1 = _inp2signal(r1)
        s_r2 = _inp2signal(r2)
        return cls(super().correlate(s_r1, s_r2))

    def __add__(self: S, a: SSPRN) -> S:
        s_a = _inp2signal_operation(a)
        return super().__add__(s_a)

    def __sub__(self: S, a: SSPRN) -> S:
        s_a = _inp2signal_operation(a)
        return super().__sub__(s_a)

    def __mul__(self: S, a: SSPRN) -> S:
        s_a = _inp2signal_operation(a)
        return super().__mul__(s_a)

    def __truediv__(self: S, a: SSPRN) -> S:
        s_a = _inp2signal_operation(a)
        return super().__truediv__(s_a)

    def __pow__(self: S, a: SSPRN) -> S:
        s_a = _inp2signal_operation(a)
        return super().__pow__(s_a)


def _inp2signal_operation(inp: SSPRN) -> Union[Relation, Signal, Number]:
    if isinstance(inp, spectrum.Spectrum):
        return inp.get_signal()
    else:
        return inp


def _inp2signal(inp: SSPR) -> Signal:
    if isinstance(inp, spectrum.Spectrum):
        return inp.get_signal()
    elif isinstance(inp, Signal):
        return inp
    elif isinstance(inp, Relation):
        return Signal(inp)
    else:
        raise ConvertingError(type(inp), Signal)



from typing import Callable, Optional, TypeVar, Union

import numpy as np

from ..relation import Relation
from ..signal import Signal
from .emd_analyze import get_IMFs_emd
from .a_t import tukey_a_t


def soft_clip(
    data: np.ndarray, limits: float, percent=0.85, coefficient: float = 1
) -> np.ndarray:
    """Custom function for correction amplitude."""
    hard_limit = limits
    linear_limit = limits * percent

    def select_choice(data: np.ndarray):

        amplitude = abs(data[0])

        if amplitude <= linear_limit:
            return np.array((data[0], data[1]))

        if amplitude >= hard_limit:
            return np.array(
                (data[0], ((hard_limit / amplitude) ** coefficient) * data[1]))

        scale = hard_limit - linear_limit
        compression = scale * \
            np.sin(np.pi / 2 * (amplitude - linear_limit) / scale)

        return np.array((data[0], (((linear_limit + compression) / amplitude)
                                   ** coefficient) * data[1]))

    return np.apply_along_axis(select_choice, 1, data)


def get_correction_for_source(
    signal: Signal,
    reaction_mass: float = 1.0,
    limits: float = None,
    limits_percent=0.85,
    limit_iteration: Optional[int] = 10,
    window_percent=0.01,
    coefficient_function: Callable[[float], float] = lambda x: x,
) -> Signal:
    '''Sweep signal correction for realization on the vibration source.

    Steps of corrections:
    1. Calculate displacement from force.
    2. Using EMD find first IMFs of displacement.
    3. Apply suppression amplitude after limits.
    4. Apply window at the start to ensure a zero first amplitude.
    5. Return calculated force from displacement

    Args:
        signal (Signal): signal to be corrected.

        reaction_mass (float, optional): reaction mass of source. Defaults to 1.0.

        limits (float, optional): displacement limitation of source. Defaults to None.

        limits_percent (float, optional): from 0 to 1, determine the limits =
            `limits*limits_percent` up to which the displacement amplitude will
            not changed, after that, the limit amplitude will be changed using the
            `soft_clip` function. Defaults to 0.85.

        limit_iteration (Optional[int], optional): iterate corrections. Defaults to 10.

        window_percent (float, optional): apply window at the initial displacement
            to ensure a zero first amplitude. Defaults to 0.01.

        coefficient_function (_type_, optional): function to suppress..
            Defaults to lambda x:x.

    Returns:
        Signal: correct force signal.
    '''

    new_time = signal.x.copy()
    new_time.end = new_time.end - 2 * new_time.sample

    window = Relation(
        new_time,
        tukey_a_t(
            new_time.array,
            new_time.end *
            window_percent,
            "left"))

    displacement = signal.integrate().integrate() / reaction_mass
    displacement = displacement.shift(-2 * displacement.sample)
    new_displacement = None

    imfs = get_IMFs_emd(displacement)

    imfs[0] = imfs[0] * window

    d_array = np.vstack((imfs[0].y, signal.y[1:-1]))
    d_array = np.transpose(d_array)

    cnt: int = 0
    while True and limits is not None:
        cnt += 1
        result = d_array

        for _ in range(cnt):
            result = soft_clip(
                result,
                limits,
                limits_percent,
                coefficient_function(cnt))

        force = type(signal)(new_time, result[:, 1])
        new_displacement = force.integrate().integrate() / reaction_mass
        imfs = get_IMFs_emd(new_displacement)

        new_displacement = imfs[0] * window

        if np.all(np.abs(new_displacement.y) < limits):
            break

        if limit_iteration is not None and cnt + 1 > limit_iteration:
            break

    new_displacement = new_displacement or imfs[0]

    return new_displacement.diff().diff() * reaction_mass



from ..defaults import methods as dfm
from ..axis import get_array_axis_from_array


class Config:
    """The configuration class.

    The methods used to calculate the various parameters are provided in the
    configuration file. Also the file contain an other parameters.

    ---

    `get_array_axis_from_array_method`:

    Method to create new instance of ArrayAxis from some array of numbers.

    Args:
        x (ArrayLike): input array_like of numbers.
        round_dx (bool, optional): if True then round sample. Defaults to True.

    Returns:
        ArrayAxis: new ArrayAxis.

    ---


    `interpolate_extrapolate_method`:

    The method by which interpolation and extrapolation are performed.
    The method returns a function that takes a new x sequence and
    return a new y sequence.

    Method derived from default function:
    sweep_design.defaults.methods.interpolate_extrapolate

    Args:
        x (X): numbers array of axis. Samples can be not equal.

        y (Y): Representation interpolated extrapolated functions
            as array.

        bounds_error (bool, optional): if False then do not raise error if new
            array behind of bound old array. Defaults to False.

        fill_value (float, optional): default fill value if other not expected.
            Defaults to 0.0.

    Returns:
        Callable[[X], Y]: Callable that get first new array of x and return
            interpolate-extrapolate result.

    ---

    `math_operation_method`:

    Method for math operation
    Method derived from default function:
    `sweep_design.defaults.math_operation`

    Args:
        y1 (np.ndarray): first sequence y.
        y2 (Union[np.ndarray, Number]): second sequence y or other number
            name_operation (MathOperation): which mathematical operation (+, -,
            \\*, / and etc.)

    Raises:
        TypeFuncError: if operation can not be executed.

    Returns:
        Y: result of math operation.

    ---

    `integrate_one_method`:

    Method for calculating the integral of a sequence on a segment.
    Method derived from default function:
    `sweep_design.defaults.methods.one_integrate`

    Args:
        relation (Relation): from will be calculated integral.

    Returns:
        float: result of integration.

    ---

    `integrate_method`:

    The method by which the integration array is performed. Integration across
    the entire function. Get the expected integrated array function.
    Method derived from default function:
    `sweep_design.defaults.methods.integrate`

    Args:
        relation (Relation): integrated function.

    Returns:
        Tuple[XAxis, Y]: result of integration of function.

    ---

    `integrate_function_method`:

    The method by which the integration function is performed. Integration across
    the entire function. Get the expected integrated array function.
    Method derived from default function:
    `sweep_design.defaults.methods.integrate_function`

    Args:
        function (Callable[[x], y]): function is describing
            changes frequency from time.

        x (np.ndarray): time array.

    Returns:
        Relation: result of integration function.

    ---

    `differentiate_method`:

    The method by which differentiation is performed.
    Method derived from default function:
    `sweep_design.defaults.methods.differentiate`

    Args:
        relation (Relation): function which will be differentiated.

    Returns:
        Tuple[XAxis, Y]: result of differentiation.

    ---

    `correlate_method`:

    The method by which the correlation is performed.
    Method derived from default function:
    `sweep_design.defaults.methods.correlate`

    Args:
        cls (Type[&quot;Relation&quot;]): cls to use equalization of two arrays.
        r1 (Relation): first function y.
        r2 (Relation): second function y.

    Returns:
        Tuple[XAxis, np.ndarray]: result of correlation.

    ---

    `convolve_method`:

    The method by which the convolution is performed.
    Method derived from default function:
    `sweep_design.defaults.methods.convolve`

    Args:
        cls (Type[&quot;Relation&quot;]): class to use equalization of two arrays.
        r1 (Relation): first function y.
        r2 (Relation): second function y.

    Returns:
        Tuple[XAxis, np.ndarray]: result of convolution.

    ---

    `get_common_x`:

    A method by which to find the common sequence of numbers along
    the x-axis, obtained from two other sequences along the x-axis.
    Method derived from default function:
    `sweep_design.defaults.methods.get_common_x`

    Args:
        x1 (XAxis): first axis.
        x2 (XAxis): second axis.

    Returns:
        XAxis: return common axis.

    ---

    `spectrum2signal_method`:

    Method for converting a spectrum into a signal. (Using Fourier transform)
    Method derived from default function:
    `sweep_design.defaults.methods.spectrum2signal`

    Args:
        relation (Relation): spectrum of signal.
        time_start (float, optional): default fft convert to 0. time. Maybe you
            want another start of time. Defaults to None.

    Returns:
        Tuple[TimeAxis, np.ndarray]: result transformation spectrogram to signal.

    ---

    `signal2spectrum_method`:
    Method for converting a signal into a spectrum. (Using Fourier transform)
    Method derived from default function:
    `sweep_design.defaults.methods.signal2spectrum`

    Args:
        relation (Relation): signal from which get spectrum.
        is_start_zero (bool, optional): Consider array started from zero time.
            Defaults to False.

    Returns:
        Tuple[FrequencyAxis, np.ndarray]: result transformation signal to
            spectrum.

    ---

    The above methods can be overridden with your own here, or you can import the
    class `Config` somewhere and override it there.
    (They must be written according to the rules corresponding to
    the **input** and **output** parameters)

    """

    # Methods for the Relation.
    get_array_axis_from_array_method = get_array_axis_from_array
    interpolate_extrapolate_method = dfm.interpolate_extrapolate
    math_operation = dfm.math_operation
    integrate_one_method = dfm.one_integrate
    integrate_method = dfm.integrate
    integrate_function_method = dfm.integrate_function
    differentiate_method = dfm.differentiate
    correlate_method = dfm.correlate
    convolve_method = dfm.convolve
    get_common_x = dfm.get_common_x

    # Methods for Spectrum and Signal.
    spectrum2signal_method = dfm.spectrum2signal
    signal2spectrum_method = dfm.signal2spectrum



from .a_t import tukey_a_t
from .emd_analyze import get_IMFs_ceemdan, get_IMFs_emd
from .f_t import f_t_linear_array, f_t_linear_function
from .ftat_functions import proportional_freq2time, dwell
from .sweep_correction import correct_sweep
from .source_sweep_correction import get_correction_for_source
